
import asyncio
from os import makedirs
from os.path import join, expanduser
from threading import Thread, Event
from time import sleep

from notebook.notebookapp import NotebookApp
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

from ..command.args import JUPYTER as J, ROOT


class JupyterState: pass


JUPYTER = JupyterState()
JUPYTER.ENABLED = False
JUPYTER.CONNECTION_URL = None
JUPYTER.NOTEBOOK_DIR = None

__RUNNING = False


def set_jupyter_args(log, args):
    global JUPYTER
    JUPYTER.ENABLED = args[J]
    JUPYTER.NOTEBOOK_DIR = expanduser(join(args[ROOT], 'notebooks'))
    log.debug(f'Creating {JUPYTER.NOTEBOOK_DIR}')
    makedirs(JUPYTER.NOTEBOOK_DIR, exist_ok=True)


def start_jupyter(log):
    global __RUNNING, JUPYTER
    if JUPYTER.ENABLED and not __RUNNING:
        started = Event()

        def start():
            log.info('Starting Jupyter server in separate thread')
            asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
            JupyterServer.launch_instance(['--notebook-dir', JUPYTER.NOTEBOOK_DIR], log=log, started=started)

        t = Thread(target=start)
        t.daemon = True
        t.start()
        started.wait()  # set in JupyterServer.start() which is as late as we can get in startup
        sleep(5)  # annoying, but we seem to need extra time or web.open() doesn't work
        __RUNNING = True
        JUPYTER.CONNECTION_URL = JupyterServer._instance.connection_url


def stop_jupyter(log):
    global __RUNNING
    if __RUNNING:
        try:
            log.info('Stopping Jupyter server')
            JupyterServer._instance.stop()
            log.debug('Jupyter server stopped')
        except Exception as e:
            log.warning(f'Error stopping JupyterState: {e}')
        finally:
            __RUNNING = False
    else:
        log.debug('Jupyter server not running')


class JupyterServer(NotebookApp):

    def __init__(self, log=None, started=None, **kwargs):
        self._log = log
        self._started = started
        super().__init__(**kwargs)

    @property
    def log_level(self):
        # this silences jupyter's logging
        # the existing mechanism only lets you set a value of 50, which means that "critical" messages
        # are still logged, and the decidedly non-critical usage instructions are printed.
        return 60

    def init_signal(self):
        self._log.debug('Skipping signal init')

    def start(self):
        self._started.set()
        super().start()




from collections import Counter
from itertools import chain
from json import loads

from scipy.interpolate import UnivariateSpline

from . import WaypointCalculator
from .climb import find_climbs, Climb
from .heart_rate import hr_zones_from_database
from ..load import StatisticJournalLoader
from ..names import ACTIVE_DISTANCE, MAX, M, ACTIVE_TIME, S, ACTIVE_SPEED, KMH, round_km, MEDIAN_KM_TIME, \
    PERCENT_IN_Z, PC, TIME_IN_Z, HR_MINUTES, MAX_MED_HR_M, BPM, MIN, CNT, SUM, AVG, MSR, summaries, HEART_RATE, \
    DISTANCE, ELEVATION, CLIMB_ELEVATION, CLIMB_DISTANCE, CLIMB_TIME, CLIMB_GRADIENT, TOTAL_CLIMB, RAW_ELEVATION
from ..waypoint import Chunks
from ...squeal import Constant, StatisticName, StatisticJournalFloat, StatisticJournalInteger


class ActivityStatistics(WaypointCalculator):

    # for historical reasons, and because it adds few stats, this still doesn't use a loader.

    def _run_activity(self, s, activity_group):
        climb = self._assert_karg('climb')
        self.__climb = Climb(**loads(Constant.get(s, climb).at(s).value))
        return super()._run_activity(s, activity_group)

    def _filter_statistic_journals(self, q):
        return q.filter(StatisticName.name == ACTIVE_TIME)

    def _names(self):
        names = {HEART_RATE: 'heart_rate',
                 DISTANCE: 'distance',
                 RAW_ELEVATION: 'raw_elevation',
                 ELEVATION: 'elevation'}
        return names

    def _add_float_stat(self, s, ajournal, name, summary, value, units, time=None):
        if time is None:
            time = ajournal.start
        StatisticJournalFloat.add(self._log, s, name, units, summary, ActivityStatistics,
                                  ajournal.activity_group, ajournal, value, time)

    def _add_int_stat(self, s, ajournal, name, summary, value, units, time=None):
        if time is None:
            time = ajournal.start
        StatisticJournalInteger.add(self._log, s, name, units, summary, ActivityStatistics,
                                    ajournal.activity_group, ajournal, int(round(value)), time)

    def _add_stats_from_waypoints(self, s, ajournal, waypoints):
        totals = self._add_totals(s, ajournal, waypoints)
        self._add_times_for_distance(s, ajournal, waypoints)
        self._add_hr_stats(s, ajournal, waypoints, totals)
        waypoints = self._fix_elevation(s, ajournal, waypoints)
        self._add_climbs(s, ajournal, waypoints)

    def _add_totals(self, s, ajournal, waypoints):
        totals = Totals(self._log, waypoints)
        self._add_float_stat(s, ajournal, ACTIVE_DISTANCE, summaries(MAX, CNT, SUM, MSR), totals.distance, M)
        self._add_float_stat(s, ajournal, ACTIVE_TIME, summaries(MAX, SUM, MSR), totals.time, S)
        self._add_float_stat(s, ajournal, ACTIVE_SPEED, summaries(MAX, AVG, MSR), totals.distance * 3.6 / totals.time, KMH)
        return totals

    def _add_times_for_distance(self, s, ajournal, waypoints):
        for target in round_km():
            times = list(sorted(TimeForDistance(self._log, waypoints, target * 1000).times()))
            if not times:
                break
            median = len(times) // 2
            self._add_float_stat(s, ajournal, MEDIAN_KM_TIME % target, summaries(MIN, MSR), times[median], S)

    def _add_hr_stats(self, s, ajournal, waypoints, totals):
        zones = hr_zones_from_database(self._log, s, ajournal.activity_group, ajournal.start)
        if zones:
            for (zone, frac) in Zones(self._log, waypoints, zones).zones:
                self._add_float_stat(s, ajournal, PERCENT_IN_Z % zone, None, 100 * frac, PC)
                self._add_float_stat(s, ajournal, TIME_IN_Z % zone, None, frac * totals.time, S)
            for target in HR_MINUTES:
                heart_rates = sorted(MedianHRForTime(self._log, waypoints, target * 60).heart_rates(), reverse=True)
                if heart_rates:
                    self._add_float_stat(s, ajournal, MAX_MED_HR_M % target, summaries(MAX, MSR), heart_rates[0], BPM)
        else:
            self._log.warning('No HR zones defined for %s or before' % ajournal.start)

    def _fix_elevation(self, s, ajournal, waypoints):
        with_elevations = [waypoint for waypoint in waypoints if waypoint.raw_elevation != None]
        if len(with_elevations) > 4:
            fixed = []
            x = [waypoint.distance for waypoint in with_elevations]
            y = [waypoint.raw_elevation for waypoint in with_elevations]
            i = 1
            while i < len(x):
                if x[i-1] >= x[i]:
                    del x[i-1], y[i-1]
                else:
                    i += 1
            loader = StatisticJournalLoader(self._log, s, self)
            # the 7 here is from eyeballing various plots compared to other values
            # it seems better to smooth along the route rather that smooth the terrain model since
            # 1 - we expect the route to be smoother than the terrain in general (roads / tracks)
            # 2 - smoothing the 2d terrain is difficult to control and can give spikes
            # 3 - we better handle errors from mismatches between terrain model and position
            #     (think hairpin bends going up a mountainside)
            # the main drawbacks are
            # 1 - speed on loading
            # 2 - no guarantee of consistency between routes (or even on the same routine retracing a path)
            spline = UnivariateSpline(x, y, s=len(with_elevations) * 7)
            for waypoint in with_elevations:
                elevation = spline(waypoint.distance)
                loader.add(ELEVATION, M, None, ajournal.activity_group, ajournal,
                           elevation, waypoint.time, StatisticJournalFloat)
                fixed.append(waypoint._replace(elevation=elevation))
            loader.load()
            return fixed
        else:
            return waypoints

    def _add_climbs(self, s, ajournal, waypoints):
        total_elevation = 0
        for lo, hi in find_climbs(waypoints, params=self.__climb):
            up = hi.elevation - lo.elevation
            along = hi.distance - lo.distance
            time = (hi.time - lo.time).total_seconds()
            self._add_float_stat(s, ajournal, CLIMB_ELEVATION, summaries(MAX, SUM, MSR), up, M, time=hi.time)
            self._add_float_stat(s, ajournal, CLIMB_DISTANCE, summaries(MAX, SUM, MSR), along, M, time=hi.time)
            self._add_float_stat(s, ajournal, CLIMB_TIME, summaries(MAX, SUM, MSR), time, S, time=hi.time)
            self._add_float_stat(s, ajournal, CLIMB_GRADIENT, summaries(MAX, MSR), 100 * up / along, PC, time=hi.time)
            total_elevation += up
        if total_elevation:
            self._add_float_stat(s, ajournal, TOTAL_CLIMB, summaries(MAX, SUM, MSR), total_elevation, M)


class TimeForDistance(Chunks):

    def __init__(self, log, waypoints, distance):
        super().__init__(log, waypoints)
        self.__distance = distance

    def times(self):
        for chunks in self.chunks():
            distance = sum(chunk.distance() for chunk in chunks)
            if distance > self.__distance:
                while chunks and distance - chunks[0].distance_delta() > self.__distance:
                    distance -= chunks[0].distance_delta()
                    self.drop_first(chunks)
                time = sum(chunk.time() for chunk in chunks)
                yield time * self.__distance / distance


class MedianHRForTime(Chunks):

    def __init__(self, log, waypoints, time, max_gap=None):
        super().__init__(log, waypoints)
        self.__time = time
        self.__max_gap = 0.01 * time if max_gap is None else max_gap
        log.debug('Will reject gaps > %ds' % int(self.__max_gap))

    def _max_gap(self, chunks):
        return max(c1[0].timespan.start - c2[0].timespan.finish
                   for c1, c2 in zip(list(chunks)[1:], chunks)).total_seconds()

    def heart_rates(self):
        for chunks in self.chunks():
            while len(chunks) > 1 and self._max_gap(chunks) > self.__max_gap:
                self._log.debug('Rejecting chunk because of gap (%ds)' % int(self._max_gap(chunks)))
                chunks.popleft()
            time = sum(chunk.time() for chunk in chunks)
            if time > self.__time:
                while chunks and time - chunks[0].time_delta() > self.__time:
                    time -= chunks[0].time_delta()
                    self.drop_first(chunks)
                heart_rates = list(sorted(chain(*(chunk.heart_rates() for chunk in chunks))))
                if heart_rates:
                    median = len(heart_rates) // 2
                    yield heart_rates[median]


class Totals(Chunks):

    def __init__(self, log, waypoints):
        super().__init__(log, waypoints)
        all_chunks = self.complete()
        self.distance = sum(chunk.distance() for chunk in all_chunks)
        self.time = sum(chunk.time() for chunk in all_chunks)


class Zones(Chunks):

    def __init__(self, log, waypoints, zones):
        super().__init__(log, waypoints)
        # this assumes record data are evenly distributed
        self.zones = []
        all_chunks = self.complete()
        counts = Counter()
        lower_limit = 0
        for zone, upper_limit in enumerate(zones):
            for chunk in all_chunks:
                for heart_rate in chunk.heart_rates():
                    if heart_rate is not None:
                        if lower_limit <= heart_rate < upper_limit:
                            counts[zone] += 1  # zero-based (incremented below)
            lower_limit = upper_limit
        total = sum(counts.values())
        if total:
            for zone in range(len(zones)):
                self.zones.append((zone + 1, counts[zone] / total))





import datetime as dt
from json import loads
from re import sub

from sqlalchemy import asc, desc
from urwid import Pile, Text, Columns

from . import Displayer
from ..calculate.impulse import ImpulseStatistics, Response
from ...lib.date import local_date_to_time
from ...lib.schedule import Schedule
from ...lib.utils import label, em, error
from ...squeal.tables.constant import Constant
from ...squeal.tables.statistic import StatisticJournal, StatisticName, TYPE_TO_JOURNAL_CLASS
from ...uweird.tui.decorators import Indent


class ImpulseDiary(Displayer):

    def _display_date(self, s, f, date):
        yield from self._display_schedule(s, f, date, schedule=Schedule('d'))

    def _display_schedule(self, s, f, date, schedule=None):

        fitness = self._assert_karg('fitness')
        fatigue = self._assert_karg('fatigue')
        rows = []
        for cols in self._single_response(s, f, date, schedule, fitness, schedule.frame_type == 'd'):
            rows.append(Columns(cols))
        for cols in self._single_response(s, f, date, schedule, fatigue, schedule.frame_type == 'd'):
            rows.append(Columns(cols))
        if rows:
            yield Pile([Text('SHRIMP'), Indent(Pile(rows))])

    def _single_response(self, s, f, date, schedule, constant_name, display_range):
        start_time = local_date_to_time(schedule.start_of_frame(date))
        finish_time = local_date_to_time(schedule.next_frame(date))
        response = Response(**loads(Constant.get(s, constant_name).at(s, start_time).value))
        start = self._read(s, response.dest_name, start_time, finish_time, asc)
        finish = self._read(s, response.dest_name, start_time, finish_time, desc)
        if start and finish and start.value != finish.value:
            lo, hi = self._range(s, response.dest_name, start, finish_time, dt.timedelta(days=90))
            if lo is not None and hi is not None:
                if display_range:
                    style = 'quintile-%d' % min(5, 1 + int(5 * (finish.value - lo.value) / (hi.value - lo.value)))
                else:
                    style = 'em'
                yield [Text(response.dest_name),
                       Text([label('Frm: '), (style, '%d' % int(start.value))]),
                       Text([label('To:  '), (style, '%d' % int(finish.value))]),
                       Text(em('increase') if start.value < finish.value else error('decrease'))]
                if display_range:
                    yield [Text([label('Over 90 days')]),
                           Text([label('Lo:  '), '%d' % int(lo.value)]),
                           Text([label('Hi:  '), '%d' % int(hi.value)]),
                           Text('')]

    def _read(self, s, name, start_time, finish_time, direcn):
        return s.query(StatisticJournal). \
            join(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticName.owner == ImpulseStatistics,
                   StatisticJournal.time >= start_time,
                   StatisticJournal.time < finish_time). \
            order_by(direcn(StatisticJournal.time)). \
            limit(1).one_or_none()

    def _range(self, s, name, value, finish_time, period):
        jtype = TYPE_TO_JOURNAL_CLASS[type(value.value)]
        start_time = finish_time - period
        q = s.query(jtype). \
            join(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticName.owner == ImpulseStatistics,
                   jtype.time >= start_time,
                   jtype.time < finish_time)
        return (q.order_by(asc(jtype.value)).limit(1).one_or_none(),
                q.order_by(desc(jtype.value)).limit(1).one_or_none())




from logging import basicConfig, getLogger, DEBUG
from sys import stdout
from tempfile import TemporaryDirectory
from time import sleep
from unittest import TestCase

from ch2.command.args import JUPYTER, ROOT
from ch2.uranus.server import set_jupyter_args, stop_jupyter
from ch2.uranus.template.compare_activities import compare_activities


class TestUranus(TestCase):

    def setUp(self):
        if not getLogger().handlers:
            basicConfig(stream=stdout, level=DEBUG)
        self._log = getLogger()

    # def test_template(self):
    #     compare_activities('2018-03-01 16:00', '2017-09-19 16:00', direct=True, group='Bike')

    # def test_display(self):
    #     with TemporaryDirectory() as dir:
    #         try:
    #             self._log.debug(f'Dir {dir}')
    #             set_jupyter_args(self._log, {JUPYTER: True, ROOT: dir})
    #             compare_activities('2018-03-01 16:00', '2017-09-19 16:00', 'Bike', log=self._log)
    #             sleep(5000)
    #         finally:
    #             stop_jupyter(self._log)



from json import dumps

from ..stoats.calculate.nearby import Nearby, NearbyStatistics
from ..squeal.database import connect
from ..squeal.tables.activity import ActivityGroup
from ..squeal.tables.constant import Constant, ValidateNamedTuple
from ..squeal.tables.pipeline import Pipeline, PipelineType
from ..squeal.tables.statistic import StatisticName, StatisticJournalType
from ..squeal.tables.topic import Topic, TopicField
from ..squeal.types import long_cls
from ..uweird.fields.topic import Integer

NEARBY_CNAME = 'Nearby'


def config(*args):
    '''
    Start here to configure the system.  Create an instance on the command line:

        log, db = config('-v', '4')
        print(c...)  todo
        ...
    '''
    ns, log, db = connect(args)
    return log, db


class Counter:

    def __init__(self, start=10, delta=10):
        self.__start = start
        self.__delta = delta
        self.__previous = None

    def __call__(self, reset=None, delta=None):
        if delta is not None:
            if delta < 1:
                raise Exception('Negative increment')
            self.__delta = delta
        if reset is None:
            if self.__previous is None:
                self.__previous = self.__start
            else:
                self.__previous += self.__delta
        else:
            if reset <= self.__previous:
                raise Exception('Sort not increasing with reset')
            else:
                self.__previous = reset
        return self.__previous


def add(s, instance):
    '''
    Add an instance to the session (and so to the database), returning the instance.
    You likely don't need this - see the more specific helpers below.

    The instance can of any class that subclasses the Field class from SQLAlchemy.
    In practice, that means most classes in the ch2.squeal.tables mdoule.
    However, only some classes make sense in the context of a configuration, and
    more specific helpers probably already exist for those.
    '''
    s.add(instance)
    return instance


def add_statistics(s, cls, sort, **kargs):
    '''
    Add a class to the statistics pipeline.

    The pipeline classes are invoked when the diary is modified and when activities are added.
    They detect new data and calculate appropriate statistics.
    See the ch2.stoats module for examples.

    The sort argument fixes the order in which the classes are instantiated and called and can
    be an integer or a callable (that returns an integer) like Counter above.

    The kargs are passed to the constructor and so can be used to customize the processing.
    '''
    return add(s, Pipeline(cls=cls, type=PipelineType.STATISTIC, sort=sort, kargs=kargs))


def add_diary(s, cls, sort, **kargs):
    '''
    Add a class to the diary pipeline.

    The pipeline classes are invoked when the diary is displyed.
    They generate display classes for activity statistics (and similar)
    See the ch2.stoats module for examples.

    The sort argument fixes the order in which the classes are instantiated and called and can
    be an integer or a callable (that returns an integer) like Counter above.

    The kargs are passed to the constructor and so can be used to customize the processing.
    '''
    return add(s, Pipeline(cls=cls, type=PipelineType.DIARY, sort=sort, kargs=kargs))


def add_monitor(s, cls, sort, **kargs):
    '''
    Add a class to the monitor pipeline.

    The pipeline classes are invoked when activities are imported from FIT files.
    They read the files and create MonitorJournal entries and associated statistics.

    The sort argument fixes the order in which the classes are instantiated and called and can
    be an integer or a callable (that returns an integer) like Counter above.

    The kargs are passed to the constructor and so can be used to customize the processing.
    '''
    return add(s, Pipeline(cls=cls, type=PipelineType.MONITOR, sort=sort, kargs=kargs))


def add_activity_group(s, name, sort, description=None):
    '''
    Add an activity type to the configuration.

    These are used to group activities (and related statistics).
    So typical entries might be for cycling, running, etc.
    '''
    return add(s, ActivityGroup(name=name, sort=sort, description=description))


def add_activities(s, cls, sort, **kargs):
    '''
    Add a class to the activities pipeline.

    The pipeline classes are invoked when activities are imported from FIT files.
    They read the files and create ActivityJournal entries and associated statistics.

    The sort argument fixes the order in which the classes are instantiated and called and can
    be an integer or a callable (that returns an integer) like Counter above.

    The kargs are passed to the constructor and so can be used to customize the processing.
    '''
    return add(s, Pipeline(cls=cls, type=PipelineType.ACTIVITY, sort=sort, kargs=kargs))


def add_constant(s, name, description=None, units=None, single=False,
                 statistic_journal_type=StatisticJournalType.INTEGER):
    '''
    Add a constant (not associated with an activity).

    Configuring a constant allows the user to supply a value later, using the `ch2 constant` command.
    This can be useful for values that don't vary often, and so aren't worth adding to the diary.
    An example is FTHR, which you will only measure occasionally, but which is needed when calculating
    activity statistics (also, FTHR can vary by activity, which is why we add a constant per activity).
    '''
    statistic_name = add(s, StatisticName(name=name, owner=Constant, constraint=None,
                                          units=units, description=description))
    return add(s, Constant(statistic_journal_type=statistic_journal_type, statistic_name=statistic_name,
                           name=name, single=single))


def add_activity_constant(s, activity_group, name, description=None, units=None, single=False,
                          statistic_journal_type=StatisticJournalType.INTEGER):
    '''
    Add a constant associated with an activity.

    Configuring a constant allows the user to supply a value later, using the `ch2 constant` command.
    This can be useful for values that don't vary often, and so aren't worth adding to the diary.
    An example is FTHR, which you will only measure occasionally, but which is needed when calculating
    activity statistics (also, FTHR can vary by activity, which is why we add a constant per activity).
    '''
    if activity_group.id is None:
        s.flush()
    statistic_name = add(s, StatisticName(name=name, owner=Constant, constraint=activity_group,
                                          units=units, description=description))
    return add(s, Constant(statistic_journal_type=statistic_journal_type, statistic_name=statistic_name,
                           name='%s.%s' % (name, activity_group.name), single=single))


def add_enum_constant(s, name, enum, constraint=None, description=None, units=None, single=False):
    '''
    Add a constant that is a JSON encoded enum.  This is validated before saving.
    '''
    statistic_name = add(s, StatisticName(name=name, owner=Constant, constraint=constraint,
                                          units=units, description=description))
    return add(s, Constant(statistic_journal_type=StatisticJournalType.TEXT, statistic_name=statistic_name,
                           name=name, single=single,
                           validate_cls=ValidateNamedTuple,
                           validate_args=[], validate_kargs={'tuple_cls': long_cls(enum)}))


def set_constant(s, constant, value, time=None, date=None):
    '''
    Set a constant value.
    '''
    constant.add_value(s, value, time=time, date=date)


def name_constant(short_name, activity_group):
    '''
    Constants typically combine a name with an activity group (because they're specific to a
    particular activity).
    '''
    return '%s.%s' % (short_name, activity_group.name)


def add_topic(s, name, sort, description=None, schedule=None):
    '''
    Add a root topic.

    Topics are displayed in the diary.
    They can be permanent, or associated with some schedule.
    They can also be associated with fields (and so with statistics).

    A root topic is usually used as a header to group related children.
    For example, 'DailyDiary' to group diary entries (notes, weight, sleep etc), or 'Plan' to group training plans.
    '''
    return add(s, Topic(name=name, sort=sort, description=description, schedule=schedule))


def add_child_topic(s, parent, name, sort, description=None, schedule=None):
    '''
    Add a child topic.

    Topics are displayed in the diary.
    They can be permanent, or associated with some schedule.
    They can also be associated with fields (and so with statistics).

    A child topic is used to add additional structrure to an existing topic.
    For example, the parent topic might be "injuries" and permanent, while children are defined for
    specific injuries with a schedule that gives start and end dates.
    '''
    return add(s, Topic(parent=parent, name=name, sort=sort, description=description, schedule=schedule))


def add_topic_field(s, topic, name, sort, description=None, units=None, summary=None,
                    display_cls=Integer, **display_kargs):
    '''
    Add a field and associated statistic to a topic entry.

    This is how the user can enter values into the diary.
    The field describes how the values are displayed in the diary.
    The statistic describes how the values are stored in the database.
    '''
    if topic.id is None:
        s.flush()
    statistic_name = add(s, StatisticName(name=name, owner=topic, constraint=topic,
                                          description=description, units=units, summary=summary))
    field = add(s, TopicField(topic=topic, sort=sort, type=display_cls.statistic_journal_type,
                              display_cls=display_cls, display_kargs=display_kargs,
                              statistic_name=statistic_name))


def add_nearby(s, sort, activity_group, constraint, latitude, longitude, border=5,
               start='1970', finish='2999', height=10, width=10, constant=NEARBY_CNAME):
    '''
    Add a pipeline task (and related constant) to find nearby activities in a given geographic
    region (specified by latitude, longitude, width and height, all in degrees).
    '''
    activity_group_constraint = str(activity_group)
    nearby_name = name_constant(constant, activity_group)
    nearby = add_enum_constant(s, nearby_name, Nearby, single=True, constraint=activity_group_constraint,
                               description='Data needed to calculate nearby activities - see Nearby enum')
    set_constant(s, nearby, dumps({'constraint': constraint, 'activity_group': activity_group.name,
                                   'border': border, 'start': start, 'finish': finish,
                                   'latitude': latitude, 'longitude': longitude,
                                   'height': height, 'width': width}))
    add_statistics(s, NearbyStatistics, sort, nearby=nearby_name)





import datetime as dt
from collections import defaultdict
from collections.abc import Mapping, Sequence

import numpy as np
import pandas as pd
from sqlalchemy import inspect, select, and_
from sqlalchemy.sql.functions import coalesce

from .names import DISTANCE_KM, SPEED_KMH, MED_SPEED_KMH, MED_HR_IMPULSE_10, MED_CADENCE, WINDOW, MIN_PERIODS, \
    ELEVATION_M, CLIMB_MS, LOG_FITNESS, LOG_FATIGUE, ACTIVE_TIME_H, ACTIVE_DISTANCE_KM, MED_POWER_W
from ..lib.data import kargs_to_attr
from ..lib.date import local_time_to_time, time_to_local_time, YMD, HMS
from ..squeal import StatisticName, StatisticJournal, StatisticJournalInteger, ActivityJournal, \
    StatisticJournalFloat, StatisticJournalText, Interval, StatisticMeasure, Source
from ..squeal.database import connect, ActivityTimespan, ActivityGroup, ActivityBookmark
from ..stoats.calculate.monitor import MonitorStatistics
from ..stoats.display.nearby import nearby_any_time
from ..stoats.names import TIMESPAN_ID, LATITUDE, LONGITUDE, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, DISTANCE, \
    ELEVATION, SPEED, HR_ZONE, HR_IMPULSE_10, ALTITUDE, CADENCE, TIME, LOCAL_TIME, FITNESS, FATIGUE, REST_HR, \
    DAILY_STEPS, ACTIVE_TIME, ACTIVE_DISTANCE, POWER
from ch2.uranus.coasting import CoastingBookmark

# because this is intended to be called from jupyter we hide the log here
# other callers can use these routines by calling set_log() first.
LOG = [None]


def df(query):
    # https://stackoverflow.com/questions/29525808/sqlalchemy-orm-conversion-to-pandas-dataframe
    return pd.read_sql(query.statement, query.session.bind)


def session(*args):
    ns, log, db = connect(args)
    set_log(log)
    return db.session()


def get_log():
    if not LOG[0]:
        raise Exception('Create session first (or call set_log from within python)')
    return LOG[0]


def set_log(log):
    if log:
        LOG[0] = log


def _add_constraint(q, attribute, value, key):
    if value is not None:
        if isinstance(value, Mapping):
            if key in value:
                q = q.filter(attribute == value[key])
        elif isinstance(value, str):
            q = q.filter(attribute == value)
        elif isinstance(value, Sequence):
            q = q.filter(attribute.in_(value))
        else:
            q = q.filter(attribute == value)
    return q


def _collect_statistics(s, names, owner=None, constraint=None, log=None):
    set_log(log)
    if not names:
        names = ['%']
    statistic_ids, statistic_names = set(), set()
    for name in names:
        q = s.query(StatisticName).filter(StatisticName.name.like(name))
        q = _add_constraint(q, StatisticName.owner, owner, name)
        q = _add_constraint(q, StatisticName.constraint, constraint, name)
        for statistic_name in q.all():
            statistic_ids.add(statistic_name.id)
            statistic_names.add(statistic_name.name)
    return statistic_names, statistic_ids


def _tables():
    return kargs_to_attr(sj=inspect(StatisticJournal).local_table,
                         sn=inspect(StatisticName).local_table,
                         sji=inspect(StatisticJournalInteger).local_table,
                         sjf=inspect(StatisticJournalFloat).local_table,
                         sjt=inspect(StatisticJournalText).local_table,
                         inv=inspect(Interval).local_table,
                         at=inspect(ActivityTimespan).local_table)


def _build_statistic_journal_query(statistic_ids, start, finish, source_ids, schedule):

    # use more efficient expression interface and exploit the fact that
    # alternative journal types will be null in an outer join.

    t = _tables()

    q = select([t.sn.c.name, t.sj.c.time, coalesce(t.sjf.c.value, t.sji.c.value, t.sjt.c.value)]). \
        select_from(t.sj.join(t.sn).outerjoin(t.sjf).outerjoin(t.sji).outerjoin(t.sjt)). \
        where(t.sn.c.id.in_(statistic_ids))
    if start:
        q = q.where(t.sj.c.time >= start)
    if finish:
        q = q.where(t.sj.c.time <= finish)
    if source_ids is not None:  # avoid testing DataFrame sequences as bools
        # int() to convert numpy types
        q = q.where(t.sj.c.source_id.in_(int(id) for id in source_ids))
    if schedule:
        q = q.join(t.inv).where(t.inv.c.schedule == schedule)
    q = q.order_by(t.sj.c.time)
    return q


class MissingData(Exception): pass


def make_pad(data, times, statistic_names, quiet=False):
    err_cnt = defaultdict(lambda: 0)

    def pad():
        nonlocal data, times
        n = len(times)
        for name in statistic_names:
            if len(data[name]) != n:
                err_cnt[name] += 1
                if err_cnt[name] <= 1 and not quiet:
                    get_log().warning('Missing %s at %s (single warning)' % (name, times[-1]))
                data[name].append(None)

    return pad


def statistics(s, *statistics, start=None, finish=None, owner=None, constraint=None, source_ids=None,
               schedule=None, quiet=False, log=None):

    set_log(log)
    statistic_names, statistic_ids = _collect_statistics(s, statistics, owner, constraint)
    q = _build_statistic_journal_query(statistic_ids, start, finish, source_ids, schedule)
    data, times = defaultdict(list), []
    pad = make_pad(data, times, statistic_names, quiet=quiet)

    for name, time, value in s.connection().execute(q):
        if times and times[-1] != time:
            pad()
        if not times or times[-1] != time:
            times.append(time)
        if len(data[name]) >= len(times):
            raise Exception('Duplicate data for %s at %s ' % (name, time) +
                            '(you may need to specify more constraints to make the query unique)')
        data[name].append(value)
    pad()
    return pd.DataFrame(data, index=times)


def resolve_activity(s, local_time=None, time=None, activity_journal_id=None,
                     activity_group_name=None, activity_group_id=None, log=None):
    '''
    If activity_journal_id is given, it is returned.

    Otherwise, specify one of (local_time, time) and one of (activity_group_name, activity_group_id).
    '''

    set_log(log)
    if activity_journal_id:
        return activity_journal_id
    if local_time:
        time = local_time_to_time(local_time)
    if not time:
        raise Exception('Specify activity_journal_id or time')
    if activity_group_name:
        activity_group_id = s.query(ActivityGroup.id).filter(ActivityGroup.name == activity_group_name).scalar()
    if not activity_group_id:
        raise Exception('Specify activity_group_id or activity_group_name')
    activity_journal_id = s.query(ActivityJournal.id). \
        filter(ActivityJournal.start <= time,
               ActivityJournal.finish >= time,
               ActivityJournal.activity_group_id == activity_group_id).scalar()
    get_log().info('Using activity_journal_id=%d' % activity_journal_id)
    return activity_journal_id


def activity_statistics(s, *statistics, owner=None, constraint=None, start=None, finish=None,
                        local_time=None, time=None, bookmarks=None, activity_journal_id=None,
                        activity_group_name=None, activity_group_id=None, with_timespan=False,
                        log=None, quiet=False):

    set_log(log)
    if bookmarks:
        if start or finish or local_time or time or activity_journal_id:
            raise Exception('Cannot use bookmarks with additional activity constraints')
        return pd.concat(_activity_statistics(s, *statistics, owner=owner, constraint=constraint,
                                              start=bookmark.start, finish=bookmark.finish,
                                              activity_journal_id=bookmark.activity_journal_id,
                                              activity_group_name=activity_group_name,
                                              activity_group_id=activity_group_id, with_timespan=with_timespan,
                                              quiet=quiet)
                         for bookmark in bookmarks)
    else:
        return _activity_statistics(s, *statistics, owner=owner, constraint=constraint, start=start, finish=finish,
                                    local_time=local_time, time=time, activity_journal_id=activity_journal_id,
                                    activity_group_name=activity_group_name, activity_group_id=activity_group_id,
                                    with_timespan=with_timespan, quiet=quiet)


def _activity_statistics(s, *statistics, owner=None, constraint=None, start=None, finish=None,
                         local_time=None, time=None, activity_journal_id=None,
                         activity_group_name=None, activity_group_id=None, with_timespan=False,
                         quiet=False):

    statistic_names, statistic_ids = _collect_statistics(s, statistics, owner=owner, constraint=constraint)
    get_log().debug('Statistics IDs %s' % statistic_ids)
    activity_journal_id = resolve_activity(s, local_time=local_time, time=time,
                                           activity_journal_id=activity_journal_id,
                                           activity_group_name=activity_group_name,
                                           activity_group_id=activity_group_id)

    t = _tables()
    q = select([t.sn.c.name, t.sj.c.time, coalesce(t.sjf.c.value, t.sji.c.value, t.sjt.c.value), t.at.c.id]). \
        select_from(t.sj.join(t.sn).outerjoin(t.sjf).outerjoin(t.sji).outerjoin(t.sjt)). \
        where(and_(t.sn.c.id.in_(statistic_ids), t.sj.c.source_id == activity_journal_id)). \
        select_from(t.at). \
        where(and_(t.at.c.start <= t.sj.c.time, t.at.c.finish >= t.sj.c.time,
                   t.at.c.activity_journal_id == activity_journal_id)). \
        order_by(t.sj.c.time)
    if start:
        q = q.where(t.sj.c.time >= start)
    if finish:
        q = q.where(t.sj.c.time < finish)
    get_log().debug(q)

    data, times = defaultdict(list), []
    pad = make_pad(data, times, statistic_names, quiet=quiet)

    for name, time, value, timespan in s.connection().execute(q):
        if times and times[-1] != time:
            pad()
        if not times or times[-1] != time:
            times.append(time)
            if with_timespan:
                data[TIMESPAN_ID].append(timespan)
        if len(data[name]) >= len(times):
            raise Exception('Duplicate data for %s at %s ' % (name, time) +
                            '(you may need to specify more constraints to make the query unique)')
        data[name].append(value)
    pad()

    return pd.DataFrame(data, index=times)


def statistic_quartiles(s, *statistics, start=None, finish=None, owner=None, constraint=None, source_ids=None,
                        schedule=None, log=None):

    set_log(log)
    statistic_names, statistic_ids = _collect_statistics(s, statistics, owner, constraint)
    q = s.query(StatisticMeasure). \
        join(StatisticJournal, StatisticMeasure.statistic_journal_id == StatisticJournal.id). \
        join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
        join(Source, StatisticJournal.source_id == Source.id). \
        filter(StatisticName.id.in_(statistic_ids)). \
        filter(StatisticMeasure.quartile != None)
    if start:
        q = q.filter(StatisticJournal.time >= start)
    if finish:
        q = q.filter(StatisticJournal.time <= finish)
    if source_ids is not None:
        q = q.filter(StatisticJournal.source_id.in_(source_ids))
    if schedule:
        q = q.join((Interval, StatisticMeasure.source_id == Interval.id)). \
            filter(Interval.schedule == schedule)
    get_log().debug(q)

    raw_data = defaultdict(lambda: defaultdict(lambda: [0] * 5))
    for measure in q.all():
        raw_data[measure.source.start][measure.statistic_journal.statistic_name.name][measure.quartile] = \
            measure.statistic_journal.value
    data, times = defaultdict(list), []
    for time in sorted(raw_data.keys()):
        times.append(time)
        sub_data = raw_data[time]
        for statistic in statistic_names:
            if statistic in sub_data:
                data[statistic].append(sub_data[statistic])
            else:
                data[statistic].append(None)

    return pd.DataFrame(data, index=times)


def std_activity_statistics(s, local_time=None, time=None, activity_journal_id=None,
                            activity_group_name=None, activity_group_id=None, log=None):

    set_log(log)
    stats = activity_statistics(s, LATITUDE, LONGITUDE, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, DISTANCE,
                                ELEVATION, SPEED, HR_ZONE, HR_IMPULSE_10, ALTITUDE, CADENCE, POWER,
                                local_time=local_time, time=time, activity_journal_id=activity_journal_id,
                                activity_group_name=activity_group_name, activity_group_id=activity_group_id,
                                with_timespan=True)

    stats[DISTANCE_KM] = stats[DISTANCE]/1000
    stats[SPEED_KMH] = stats[SPEED] * 3.6
    stats[MED_SPEED_KMH] = stats[SPEED].rolling(WINDOW, min_periods=MIN_PERIODS).median() * 3.6
    stats[MED_HR_IMPULSE_10] = stats[HR_IMPULSE_10].rolling(WINDOW, min_periods=MIN_PERIODS).median()
    stats[MED_CADENCE] = stats[CADENCE].rolling(WINDOW, min_periods=MIN_PERIODS).median()
    stats[MED_POWER_W] = stats[POWER].rolling(WINDOW, min_periods=MIN_PERIODS).median().clip(lower=0)
    stats.rename(columns={ELEVATION: ELEVATION_M}, inplace=True)

    stats['keep'] = pd.notna(stats[HR_IMPULSE_10])
    stats.interpolate(method='time', inplace=True)
    stats = stats.loc[stats['keep'] == True]

    stats[CLIMB_MS] = stats[ELEVATION_M].diff() * 0.1
    stats[TIME] = pd.to_datetime(stats.index)
    stats[LOCAL_TIME] = stats[TIME].apply(lambda x: time_to_local_time(x.to_pydatetime(), HMS))

    return stats


def std_health_statistics(s, start=None, finish=None):

    # this assumes FF cover all the dates and HR/steps fit into them.  may not be true in all cases?
    # also, we downsample the FF data to hourly intervals then shift daily data to match one of those times
    # this avoids introducing gaps in the FF data when merging that mess up the continuity of the plots.
    stats_1 = statistics(s, FITNESS, FATIGUE, start=start, finish=finish).resample('1h').mean()
    stats_2 = statistics(s, REST_HR, start=start, finish=finish, owner=MonitorStatistics). \
        reindex(stats_1.index, method='nearest', tolerance=dt.timedelta(minutes=30))
    stats_3 = statistics(s, DAILY_STEPS, ACTIVE_TIME, ACTIVE_DISTANCE, start=start, finish=finish). \
        reindex(stats_1.index, method='nearest', tolerance=dt.timedelta(minutes=30))
    stats = stats_1.merge(stats_2, how='outer', left_index=True, right_index=True)
    stats = stats.merge(stats_3, how='outer', left_index=True, right_index=True)
    stats[LOG_FITNESS] = np.log10(stats[FITNESS])
    stats[LOG_FATIGUE] = np.log10(stats[FATIGUE])
    stats[ACTIVE_TIME_H] = stats[ACTIVE_TIME] / 3600
    stats[ACTIVE_DISTANCE_KM] = stats[ACTIVE_DISTANCE] / 1000
    stats[TIME] = pd.to_datetime(stats.index)
    stats[LOCAL_TIME] = stats[TIME].apply(lambda x: time_to_local_time(x.to_pydatetime(), YMD))

    return stats


def nearby_activities(s, local_time=None, time=None, activity_journal_id=None,
                      activity_group_name=None, activity_group_id=None):
    activity_journal_id = resolve_activity(s, local_time, time, activity_journal_id,
                                           activity_group_name=activity_group_name,
                                           activity_group_id=activity_group_id)
    return nearby_any_time(s, ActivityJournal.from_id(s, activity_journal_id))


def bookmarks(s, constraint, owner=CoastingBookmark):
    yield from s.query(ActivityBookmark). \
        filter(ActivityBookmark.owner == owner,
               ActivityBookmark.constraint == constraint).all()






from random import uniform
from unittest import TestCase

from ch2.lib.data import MutableAttr
from ch2.stoats.calculate.climb import find_climbs


class TestClimb(TestCase):

    def build_climb(self, xys, speed=10, noise=0, dt=1):
        t0 = 0
        x0, y0 = xys[0]
        t, x, y = t0, x0, y0
        for x1, y1 in xys[1:]:
            t1 = x1 / speed
            while t <= t1:
                x = x0 + (x1 - x0) * (t - t0) / (t1 - t0)
                y = y0 + (y1 - y0) * (t - t0) / (t1 - t0)
                yield MutableAttr({'time': t, 'elevation': y + uniform(-noise, noise), 'distance': x})
                t += dt
            t0, x0, y0 = t1, x1, y1

    def test_build(self):
        waypoints = list(self.build_climb([(0, 0), (1000, 100), (2000, 0)]))
        # print(waypoints)
        self.assertEqual(len(waypoints), 2000 / 10 + 1)
        self.assertEqual(waypoints[0].distance, 0)
        self.assertEqual(waypoints[0].time, 0)
        self.assertEqual(waypoints[0].elevation, 0)
        self.assertEqual(waypoints[-1].distance, 2000)
        self.assertEqual(waypoints[-1].time, 200)
        self.assertEqual(waypoints[-1].elevation, 0)

    def test_single(self):
        waypoints = list(self.build_climb([(0, 0), (1000, 100), (2000, 0)]))
        # print(waypoints)
        c = list(find_climbs(waypoints))
        self.assertEqual(len(c), 1)
        self.assertEqual(c[0][0].time, 0)
        self.assertEqual(c[0][0].distance, 0)
        self.assertEqual(c[0][0].elevation, 0)
        self.assertEqual(c[0][1].time, 100)
        self.assertEqual(c[0][1].distance, 1000)
        self.assertEqual(c[0][1].elevation, 100)

    def test_noisy_single(self):
        for _ in range(10):
            # increase distance to peak to avoid cutoff at 1km
            waypoints = list(self.build_climb([(0, 0), (1100, 100), (2000, 0)], noise=1))
            # print(waypoints)
            c = list(find_climbs(waypoints))
            self.assertEqual(len(c), 1)
            self.assertAlmostEqual(c[0][0].time, 0, delta=5)
            self.assertAlmostEqual(c[0][0].distance, 0, delta=50)
            self.assertAlmostEqual(c[0][0].elevation, 0, delta=5)
            self.assertAlmostEqual(c[0][1].time, 110, delta=5)
            self.assertAlmostEqual(c[0][1].distance, 1100, delta=50)
            self.assertAlmostEqual(c[0][1].elevation, 100, delta=5)

    def test_multiple(self):
        waypoints = list(self.build_climb([(0, 0), (1100, 100), (1200, 90), (1500, 150)]))
        # print(waypoints)
        c = list(find_climbs(waypoints))
        # print(c)
        self.assertEqual(len(c), 1)
        self.assertEqual(c[0][1].elevation - c[0][0].elevation, 150)
        waypoints = list(self.build_climb([(0, 0), (1100, 100), (1200, 80), (1500, 150)]))
        # print(waypoints)
        c = list(find_climbs(waypoints))
        # print(c)
        self.assertEqual(len(c), 1)
        self.assertEqual(c[0][1].elevation - c[0][0].elevation, 100)
        waypoints = list(self.build_climb([(0, 0), (1100, 100), (1200, 80), (1500, 170)]))
        # print(waypoints)
        c = list(find_climbs(waypoints))
        # print(c)
        self.assertEqual(len(c), 2)
        self.assertEqual(c[0][1].elevation - c[0][0].elevation, 100)
        self.assertEqual(c[1][1].elevation - c[1][0].elevation, 90)


    # def test_noisy_single_to_failure(self):
    #     while True:
    #         waypoints = list(self.build_climb([(0, 0), (1100, 100), (2000, 0)], noise=1))
    #         c = list(climbs(waypoints))
    #         if len(c) != 1:
    #             climbs(waypoints)  # debug breakpoint here





from collections import defaultdict

from sqlalchemy import func

from .waypoint import make_waypoint
from ..squeal.tables.constant import SystemConstant
from ..squeal.tables.statistic import StatisticJournal, StatisticName
from ..squeal.types import short_cls


class StatisticJournalLoader:
    '''
    Fast loading that requires locked access to the database (we pre-generate keys).
    '''

    def __init__(self, log, s, owner, add_serial=True):
        self._log = log
        self._s = s
        self._owner = owner
        self.__statistic_name_cache = dict()
        self.__staging = defaultdict(lambda: [])
        self.__latest = dict()
        self.__add_serial = add_serial
        self.__last_time = None
        self.__serial = 0 if add_serial else None

    def __bool__(self):
        return bool(self.__staging)

    def load(self):
        SystemConstant.acquire_lock(self._s, self)  # includes flush so query below OK
        try:
            rowid = self._s.query(func.max(StatisticJournal.id)).scalar() + 1
            for type in self.__staging:
                self._log.debug('Loading %d values for type %s' % (len(self.__staging[type]), short_cls(type)))
                for sjournal in self.__staging[type]:
                    sjournal.id = rowid
                    rowid += 1
                self._s.bulk_save_objects(self.__staging[type])
        finally:
            SystemConstant.release_lock(self._s, self)

    def add(self, name, units, summary, constraint, source, value, time, type):

        if self.__add_serial:
            if self.__last_time is None:
                self.__last_time = time
            elif time > self.__last_time:
                self.__last_time = time
                self.__serial += 1
            elif time < self.__last_time:
                raise Exception('Time travel!')

        key = (name, constraint)
        if key not in self.__statistic_name_cache:
            self.__statistic_name_cache[key] = \
                StatisticName.add_if_missing(self._log, self._s, name, units, summary, self._owner, constraint)
        if not self.__statistic_name_cache.get(key, None):
            raise Exception('Failed to get StatisticName for %s' % key)
        if not self.__statistic_name_cache[key].id:
            self._s.flush()
            if not self.__statistic_name_cache[key].id:
                raise Exception('Could not get StatisticName.id for %s' % key)

        instance = type(statistic_name_id=self.__statistic_name_cache[key].id, source_id=source.id,
                        value=value, time=time, serial=self.__serial)
        self.__staging[type].append(instance)
        if key in self.__latest:
            prev = self.__latest[key]
            if instance.time > prev.time:
                self.__latest[key] = instance
        else:
            self.__latest[key] = instance

    def latest(self, name, constraint, instance=None):
        latest = self.__latest.get((name, constraint))
        if latest:
            if instance and instance.time > latest.time:
                latest = instance
        else:
            latest = instance
        return latest

    def as_waypoints(self, names):
        Waypoint = make_waypoint(names.values())
        time_to_waypoint = defaultdict(lambda: Waypoint())
        statistic_ids = dict((name.id, name) for name in self.__statistic_name_cache.values())
        for type in self.__staging:
            for sjournal in self.__staging[type]:
                name = statistic_ids[sjournal.statistic_name_id].name
                if name in names:
                    time_to_waypoint[sjournal.time] = \
                        time_to_waypoint[sjournal.time]._replace(**{'time': sjournal.time,
                                                                    names[name]: sjournal.value})
        return [time_to_waypoint[time] for time in sorted(time_to_waypoint.keys())]




from glob import glob
from logging import getLogger, basicConfig, INFO
from os.path import basename, join, exists
from sys import stdout
from unittest import TestCase

from ch2.command.args import FIELDS, TABLES, GREP
from ch2.fit.format.read import filtered_records
from ch2.fit.format.records import no_names, append_units, no_bad_values, fix_degrees, chain, no_units
from ch2.fit.profile.fields import DynamicField
from ch2.fit.profile.profile import read_external_profile, read_fit
from ch2.fit.summary import summarize, summarize_csv, summarize_tables
from ch2.lib.tests import OutputMixin, HEX_ADDRESS, EXC_HDR_CHK, sub_extn, EXC_FLD, sub_dir, RNM_UNKNOWN, ROUND_DISTANCE


class TestFit(TestCase, OutputMixin):

    def setUp(self):
        if not getLogger().handlers:
            basicConfig(stream=stdout, level=INFO)
        self.log = getLogger()
        self.test_dir = 'data/test'
        self.profile_path = 'data/sdk/Profile.xlsx'

    def test_profile(self):
        nlog, types, messages = read_external_profile(self.log, self.profile_path)
        cen = types.profile_to_type('carry_exercise_name')
        self.assertEqual(cen.profile_to_internal('farmers_walk'), 1)
        session = messages.profile_to_message('session')
        field = session.profile_to_field('total_cycles')
        self.assertIsInstance(field, DynamicField)
        for name in field.references:
            self.assertEqual(name, 'sport')
        workout_step = messages.profile_to_message('workout_step')
        field = workout_step.number_to_field(4)
        self.assertEqual(field.name, 'target_value')
        fields = ','.join(sorted(field.references))
        self.assertEqual(fields, 'duration_type,target_type')

    def test_decode(self):
        types, messages, records = \
            filtered_records(self.log, read_fit(self.log, join(self.test_dir, 'source/personal/2018-07-26-rec.fit')),
                             profile_path=self.profile_path)
        with self.assertTextMatch(join(self.test_dir, 'target/personal/TestFit.test_decode'),
                                  filters=[HEX_ADDRESS]) as output:
            for _, _, record in records:
                print(record.into(tuple, filter=chain(no_names, append_units, no_bad_values, fix_degrees)),
                      file=output)

    def test_dump(self):
        with self.assertTextMatch(join(self.test_dir, 'target/personal/TestFit.test_dump')) as output:
            summarize(self.log, FIELDS, read_fit(self.log, join(self.test_dir, 'source/personal/2018-07-30-rec.fit')),
                      profile_path=self.profile_path, width=80, output=output)

    def test_developer(self):
        with self.assertTextMatch(join(self.test_dir, 'target/sdk/TestFit.test_developer')) as output:
            summarize(self.log, FIELDS, read_fit(self.log, join(self.test_dir, 'source/sdk/DeveloperData.fit')),
                      profile_path=self.profile_path, width=80, output=output)

    def test_personal(self):
        for fit_file in glob(join(self.test_dir, 'source/personal/*.fit')):
            file_name = basename(fit_file)
            with self.assertTextMatch(
                    join(self.test_dir, 'target/personal/TestFit.test_personal:' + file_name)) as output:
                summarize_tables(self.log, read_fit(self.log, fit_file), width=80, output=output,
                                 profile_path=self.profile_path)

    def test_timestamp_16(self):
        types, messages, records = \
            filtered_records(self.log,
                             read_fit(self.log,
                                      join(self.test_dir, 'source/personal/andrew@acooke.org_24755630065.fit')),
                             profile_path=self.profile_path)
        with self.assertTextMatch(join(self.test_dir, 'target/personal/TestFit.test_timestamp_16'),
                                  filters=[HEX_ADDRESS]) as output:
            for _, _, record in records:
                if record.name == 'monitoring':
                    print(record.into(tuple, filter=chain(no_names, append_units, no_bad_values, fix_degrees)),
                          file=output)

    def standard_csv(self, fit_path, csv_path, filters=None):
        if filters is None: filters = []
        if EXC_HDR_CHK not in filters: filters = [EXC_HDR_CHK] + filters
        with self.assertCSVMatch(csv_path, filters=filters) as output:
            summarize_csv(self.log, read_fit(self.log, fit_path), profile_path=self.profile_path,
                          warn=True, output=output, internal=True)

    def standard_csv_dir(self, dir, fit_pattern, exclude=None, filters=None):
        for source_fit_path in glob(join(self.test_dir, 'source', dir, fit_pattern)):
            if not exclude or basename(source_fit_path) not in exclude:
                source_csv_path = sub_extn(source_fit_path, 'csv')
                if exists(source_csv_path):
                    self.standard_csv(source_fit_path, source_csv_path, filters=filters)
                else:
                    self.log.warning('Could not find %s' % source_csv_path)
                target_csv_path = sub_dir(source_csv_path, 'target', 2)
                if exists(target_csv_path):
                    self.standard_csv(source_fit_path, target_csv_path, filters=filters)
                else:
                    self.log.warning('Could not find %s' % target_csv_path)

    def standard_dmp(self, source, target, format, filters=None):
        with self.assertTextMatch(target, filters=filters) as output:
            summarize(self.log, format, read_fit(self.log, source),
                      warn=True, profile_path=self.profile_path, output=output)

    def standard_dmp_dir(self, dir, fit_pattern, format, exclude=None, filters=None):
        for source_fit_path in glob(join(self.test_dir, 'source', dir, fit_pattern)):
            if not exclude or basename(source_fit_path) not in exclude:
                target_dmp_path = sub_extn(sub_dir(source_fit_path, 'target', 2), format[:3])
                self.standard_dmp(source_fit_path, target_dmp_path, format, filters=filters)

    def test_sdk_csv(self):
        self.standard_csv_dir('sdk', '*.fit', exclude='Activity.fit')
        # afaict it should be 0, which is mapped by the type.  the value in the CSV makes no sense.
        self.standard_csv_dir('sdk', 'Activity.fit', filters=[EXC_FLD('timer_trigger')])

    def test_personal_csv(self):
        self.standard_csv_dir('personal', '*.fit', filters=[RNM_UNKNOWN])

    # these are tests only for consistency across releases, not correctness
    def test_personal_tab(self):
        self.standard_dmp_dir('personal', '*.fit', TABLES)

    # these depend on the output from test_fix_fit, so will fail if those change
    def test_pyfitparse_tab(self):
        self.standard_dmp_dir('python-fitparse', '*.fit', TABLES,
                              exclude=['activity-unexpected-eof.fit',  # data size incorrect
                                       'activity-settings-nodata.fit',   # data size incorrect
                                       'activity-filecrc.fit',  # bad checksum
                                       'activity-activity-filecrc.fit',  # data size incorrect
                                       'activity-settings-corruptheader.fit',  # data size incorrect
                                       'event_timestamp.fit',  # data size incorrect
                                       'activity-settings.fit',  # data size incorrect
                                       ])

    # these are tests only for consistency across releases, not correctness
    def test_pyfitparse_fix_tab(self):
        self.standard_dmp_dir('python-fitparse-fix', '*.fit', TABLES)

    def test_pyfitparse_csv(self):
        self.standard_csv_dir('python-fitparse', '*.fit', filters=[RNM_UNKNOWN, ROUND_DISTANCE])

    def test_other_csv(self):
        self.standard_csv_dir('other', '*.fit', filters=[RNM_UNKNOWN, ROUND_DISTANCE])

    def test_grep(self):
        data = read_fit(self.log, join(self.test_dir, 'source/personal/2018-07-26-rec.fit'))
        summarize(self.log, GREP, data, grep=['.*:.*speed>10'])

    def test_python(self):
        # for other-projects.md
        from logging import basicConfig, getLogger, INFO
        from ch2.fit.profile.profile import read_fit, read_profile
        from ch2.fit.format.records import fix_degrees, no_units
        from ch2.fit.format.read import parse_data

        basicConfig(level=INFO)
        log = getLogger()

        data = read_fit(log, 'data/test/source/personal/2018-07-26-rec.fit')
        types, messages = read_profile(log)
        state, tokens = parse_data(log, data, types, messages)

        LAT, LONG = 'position_lat', 'position_long'
        positions = []

        for offset, token in tokens:
            record = token.parse_token().as_dict(no_units, fix_degrees)
            if record.name == 'record':
                positions.append((record.data[LAT][0], record.data[LONG][0]))

        print('Read %s positions' % len(positions))

        self.assertAlmostEqual(positions[0][0], -33.42, places=1)
        self.assertAlmostEqual(positions[0][1], -70.61, places=1)




import datetime as dt
from collections import namedtuple
from json import loads
from math import exp

from sqlalchemy import desc, inspect, select, and_

from . import IntervalCalculator
from .heart_rate import HRImpulse
from ..load import StatisticJournalLoader
from ..names import MAX
from ...lib.date import local_date_to_time
from ...lib.schedule import Schedule
from ...squeal import Constant, Interval, StatisticJournal, StatisticName, StatisticJournalFloat
from ...squeal.database import add
from ...squeal.types import short_cls

# constraint comes from constant
Response = namedtuple('Response', 'src_name, src_owner, dest_name, tau_days, scale, start')


class ImpulseStatistics(IntervalCalculator):

    def _on_init(self, *args, **kargs):
        kargs['schedule'] = 'm'
        super()._on_init(*args, **kargs)

    def _filter_intervals(self, q):
        return q.filter(Interval.schedule == self._kargs.schedule,
                        Interval.owner == self)

    def _previous(self, s, dest, start):
        return s.query(StatisticJournal). \
            join(StatisticName). \
            filter(StatisticName.name == dest,
                   StatisticJournal.time < local_date_to_time(start)). \
            order_by(desc(StatisticJournal.time)).limit(1).one_or_none()

    def _impulses(self, s, source, interval):

        sj = inspect(StatisticJournal).local_table
        sjf = inspect(StatisticJournalFloat).local_table

        stmt = select([sj.c.time, sjf.c.value]). \
            select_from(sj.join(sjf)). \
            where(and_(sj.c.statistic_name_id == source.id,
                       sj.c.time >= local_date_to_time(interval.start),
                       sj.c.time < local_date_to_time(interval.finish))). \
            order_by(desc(sj.c.time))
        # self._log.debug(stmt)
        return s.connection().execute(stmt)

    def _decay(self, response, value, time):
        dt = (time - value[0]).total_seconds()
        return time, value[1] * exp(-dt / (response.tau_days * 24 * 60 * 60))

    def _hours(self, interval):
        day = interval.start
        while day < interval.finish:
            start = local_date_to_time(day)
            for hours in range(24):
                yield start + dt.timedelta(hours=hours), None
            day += dt.timedelta(days=1)

    def _add_response(self, loader, s, response, interval, source):
        prev = loader.latest(response.dest_name, source.constraint,
                             self._previous(s, response.dest_name, interval.start))
        if prev:
            value = (prev.time, prev.value)
        else:
            value = None
        impulses = list(self._impulses(s, source, interval))
        days = list(self._hours(interval))
        prev_time = None
        # move hours very slightly later in sort so that we can de-duplicate times
        for time, impulse in sorted(impulses + days,
                                    key=lambda x: x[0] +
                                                  (dt.timedelta(seconds=0.01) if x[1] is None else dt.timedelta())):
            if prev_time is None or time != prev_time:
                if impulse:
                    if not value:
                        value = (time, response.start)
                    else:
                        value = self._decay(response, value, time)
                    value = (time, value[1] + response.scale * impulse)
                    loader.add(response.dest_name, None, MAX, source.constraint, interval, value[1], value[0],
                               StatisticJournalFloat)
                else:
                    if value:
                        value = self._decay(response, value, time)
                        loader.add(response.dest_name, None, MAX, source.constraint, interval, value[1], value[0],
                                   StatisticJournalFloat)
            prev_time = time

    def _run_calculations(self, schedule):

        from .heart_rate import HeartRateStatistics

        impulse = self._assert_karg('impulse')
        responses = self._assert_karg('responses')
        schedule = Schedule(self._kargs.schedule)

        with self._db.session_context() as s:

            constants = [Constant.get(s, response) for response in responses]
            responses = [Response(**loads(constant.at(s).value)) for constant in constants]
            impulse = HRImpulse(**loads(Constant.get(s, impulse).at(s).value))
            start, finish = Interval.first_missing_date(self._log, s, schedule, self)

            if start:
                loader = StatisticJournalLoader(self._log, s, self, add_serial=False)
                start = schedule.start_of_frame(start)
                # delete forwards
                Interval.clean_dates(self._log, s, start, finish, owner=self)
                while start <= finish:
                    interval = add(s, Interval(start=start, finish=schedule.next_frame(start),
                                               schedule=schedule, owner=self))
                    self._log.info('Running %s for %s' % (short_cls(self), interval))
                    for response, constant in zip(responses, constants):
                        activity_group = constant.statistic_name.constraint
                        source = s.query(StatisticName). \
                            filter(StatisticName.name == impulse.dest_name,
                                   StatisticName.owner == HeartRateStatistics,
                                   StatisticName.constraint == activity_group).one_or_none()
                        if source:  # None if no data loaded
                            self._add_response(loader, s, response, interval, source)
                        else:
                            self._log.warning('No values for %s' % impulse.dest_name)
                    start = schedule.next_frame(start)
                loader.load()




from contextlib import contextmanager
from time import time

from sqlalchemy import Column, Integer, UniqueConstraint, func

from ..types import Time, ShortCls, Str
from ..support import Base


class Timestamp(Base):

    '''
    This was introduced to simplify tracking dependencies for statistics.

    For example, we only need to re-calculate summary statistics for data that are newer than the latest
    summary.

    Previously, dependencies were managed by complex logic (outer joins with missing values) that were
    hard to write and hard to understand later.  Also, they had the following problems:

    * Trying to re-calculate heart rate statistics when the "missing" values are caused by the activity
      having no HRM data.

    * Nearby activities being restricted to geographic regions, making it hard to decide if all relevant
      activities were included.
    '''

    __tablename__ = 'timestamp'

    id = Column(Integer, primary_key=True)
    time = Column(Time, nullable=False, default=time)
    owner = Column(ShortCls, nullable=False)
    constraint = Column(Str)
    key = Column(Integer)
    UniqueConstraint(owner, constraint, key)

    @classmethod
    def set(cls, s, owner, constraint=None, key=None):
        cls.clear(s, owner, constraint=constraint, key=key)
        s.add(Timestamp(owner=owner, constraint=constraint, key=key))

    @classmethod
    def clear(cls, s, owner, constraint=None, key=None):
        exists = s.query(Timestamp). \
            filter(Timestamp.owner == owner,
                   Timestamp.constraint == constraint,
                   Timestamp.key == key).one_or_none()
        if exists:
            s.delete(exists)

    @classmethod
    def clear_after(cls, s, time, owner, constraint=None):
        q = s.query(Timestamp). \
            filter(Timestamp.owner == owner,
                   Timestamp.constraint == constraint)
        if time:
            q = q.filter(Timestamp.time >= time)
        q.delete()

    @contextmanager
    def on_success(self, s):
        self.clear(s, self.owner, constraint=self.constraint, key=self.key)
        yield None
        self.set(s, self.owner, constraint=self.constraint, key=self.key)




from unittest import TestCase

from ch2.lib.data import MutableAttr


class TestData(TestCase):

    def test_attr(self):
        d = MutableAttr()
        d['foo'] = 'bar'
        self.assertTrue('foo' in d)
        self.assertEqual(d.foo, 'bar')





from tempfile import NamedTemporaryFile
from unittest import TestCase

from sqlalchemy.sql.functions import count

from ch2.command.args import bootstrap_file, m, V, DEV, mm
from ch2.command.constants import constants
from ch2.config.default import default
from ch2.squeal.tables.constant import Constant


class TestConstant(TestCase):

    def test_constant(self):
        with NamedTemporaryFile() as f:
            args, log, db = bootstrap_file(f, m(V), '5')
            bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)
            with db.session_context() as s:
                n = s.query(count(Constant.id)).scalar()
                self.assertEqual(n, 8)
            args, log, db = bootstrap_file(f, m(V), '5', 'constants', '--set', 'FTHR.%', '154')
            constants(args, log, db)
            with db.session_context() as s:
                n = s.query(count(Constant.id)).scalar()
                self.assertEqual(n, 8)
                # todo - maybe test for value?




from .tables.activity import ActivityGroup, ActivityTimespan, ActivityJournal, ActivityBookmark
from .tables.constant import SystemConstant, Constant
from .tables.monitor import MonitorJournal
from .tables.nearby import ActivitySimilarity, ActivityNearby
from .tables.pipeline import Pipeline, PipelineType
from .tables.segment import Segment, SegmentJournal
from .tables.source import Source, Interval, NoStatistics
from .tables.statistic import StatisticName, StatisticJournalFloat, StatisticJournalText, StatisticJournalInteger, \
    StatisticJournal, StatisticMeasure
from .tables.topic import TopicJournal, Topic
from .tables.timestamp import Timestamp



from .focus import OnFocus, FocusAttr


class Factory:

    def __init__(self, tabs=None, bar=None):
        self.tabs = tabs
        self.bar = bar

    def __call__(self, widget, message=None):
        if self.bar:
            if message:
                widget = OnFocus(widget, message, self.bar)
        elif message:
            raise Exception('Message but no bar for %s (type %s)' % (widget, type(widget)))
        if self.tabs is not None:
            widget = self.tabs.append(widget)
        else:
            widget = FocusAttr(widget)
        return widget




from ch2 import main

main()

# import cProfile, pstats, io
# pr = cProfile.Profile()
# pr.enable()
# main()
# pr.disable()
# s = io.StringIO()
# sortby = 'cumulative'
# ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
# ps.print_stats()
# print(s.getvalue())




from sys import stdout, stderr

from .args import PATH, DROP, OUTPUT, SLICES, RAW, WARN, MIN_SYNC_CNT, MAX_RECORD_LEN, MAX_DROP_CNT, MAX_BACK_CNT, \
    MAX_FWD_LEN, DISCARD, FORCE, VALIDATE, ADD_HEADER, HEADER_SIZE, PROTOCOL_VERSION, PROFILE_VERSION, MAX_DELTA_T, \
    NAME, FIX_HEADER, FIX_CHECKSUM, NAME_BAD, NAME_GOOD, mm, no, START
from ..fit.fix import fix
from ..fit.profile.profile import read_fit


def fix_fit(args, log, db):
    '''
## fix-fit

    > ch2 fix-fit PATH -o PATH --drop

Try to fix a corrupted fit file.

If `--header` is specified then a new header is prepended at the start of the data.

If `--slices` is specified then the given slices are taken from the data and used to construct a new file.

If `--drop` is specified then the program tries to find appropriate slices by discarding data until all the
remaining data can be parsed.

If `--fix-header` is specified then the header is corercted.

If `--fix-checksum` is specified then the checksum is corrected.

### Examples

    > ch2 fix-fit FILE.FIT --slices 1000: --fix-header --fix-checksum

Will attempt to drop the first 1000 bytes from the given file.

    > ch2 fix-fit data/tests/personal/8CS90646.FIT --drop --fix-header --fix-checksum

Will attempt to fix the given file (in the test data from git).

    > ch2 fix-fit FILE.FIT --add-header --header-size 14 --slices :14,28: --fix-header --fix-checksum

Will prepend a new 14 byte header, drop the old 14 byte header, and fix the header and checksum values.
    '''

    check = args[NAME] is not None
    if check:
        name = NAME_GOOD if args[NAME] else NAME_BAD
        if args[ADD_HEADER] or args[DROP] or args[SLICES] or args[START] or args[FIX_HEADER] or args[FIX_CHECKSUM]:
            raise Exception('Cannot check (%s) and modify at the same time' % mm(name))
        if not args[VALIDATE]:
            raise Exception('%s and %s makes no sense, numpty' % (mm(name), no(VALIDATE)))
    if not args[FORCE]:
        log.warning('%s means that data are not completely parsed' % no(FORCE))

    for path in args[PATH]:

        log.info('Input ----------')
        log.info('Reading binary data from %s' % path)
        data = read_fit(log, path)
        log.debug('Read %d bytes' % len(data))

        try:
            data = fix(log, data, warn=args[WARN],
                       add_header=args[ADD_HEADER], drop=args[DROP], slices=args[SLICES], start=args[START],
                       fix_header=args[FIX_HEADER], fix_checksum=args[FIX_CHECKSUM],
                       force=args[FORCE], validate=args[VALIDATE],
                       header_size=args[HEADER_SIZE], protocol_version=args[PROTOCOL_VERSION],
                       profile_version=args[PROFILE_VERSION], min_sync_cnt=args[MIN_SYNC_CNT],
                       max_record_len=args[MAX_RECORD_LEN], max_drop_cnt=args[MAX_DROP_CNT],
                       max_back_cnt=args[MAX_BACK_CNT], max_fwd_len=args[MAX_FWD_LEN], max_delta_t=args[MAX_DELTA_T])
        except:
            if check:
                if not args[NAME]:
                    print(path)
            else:
                raise
        else:
            if check and args[NAME]:
                print(path)

        if check:
            # interleave logging and names
            stderr.flush()
            stdout.flush()
        else:
            log.info('Output ----------')
            if args[DISCARD]:
                log.info('Discarded output')
            else:
                out_path = args[OUTPUT]
                if out_path:
                    log.info('Writing binary data to %s' % out_path)
                    with open(out_path, 'wb') as out:
                        out.write(data)
                elif args[RAW]:
                    log.info('Writing binary data to stdout')
                    stdout.buffer.write(data)
                else:
                    log.info('Writing hex data to stdout')
                    stdout.write(data.hex())
                log.debug('Wrote %d bytes' % len(data))





from json import dumps

from ..stoats.calculate.power import Power
from . import name_constant, add_enum_constant, set_constant

POWER_CNAME = 'Power'


def add_power(s, activity_group, cda=0.44, crr=0, m=64+12, p=1.225, g=9.8):
    '''
    Add the constants necessary to estimate power output.
    '''
    activity_group_constraint = str(activity_group)
    power_name = name_constant(POWER_CNAME, activity_group)
    power = add_enum_constant(s, power_name, Power, single=False, constraint=activity_group_constraint,
                              description='Data needed to estimate power - see Power enum')
    set_constant(s, power, dumps({'cda': cda, 'crr': crr, 'm': m, 'p': p, 'g': g}))




import datetime as dt
from collections.abc import MutableMapping
from functools import reduce
from operator import and_

from urwid import connect_signal


def or_none(f): return lambda x: x if x is None else f(x)


# transforms
DATE_ORDINAL = (or_none(dt.date.fromordinal), or_none(lambda x: x.toordinal()))


class TransformedView(MutableMapping):
    """
    Provide a transformed view of the data, with types more
    that match the database (eg exchange NULL and empty string,
    or convert ordinals to dates).
    """

    def __init__(self, data, transforms=None):
        self.__data = data
        if transforms is None: transforms = {}
        self.__transforms = transforms

    def set_transforms(self, name, transforms):
        self.__transforms[name] = transforms

    def __getitem__(self, name):
        value = self.__data[name]
        if name in self.__transforms:
            value = self.__transforms[name][1](value)
        return value

    def __setitem__(self, name, value):
        if name in self.__transforms:
            value = self.__transforms[name][0](value)
        self.__data[name] = value

    def __delitem__(self, name):
        raise NotImplemented()

    def __len__(self):
        return len(self.__data)

    def __iter__(self):
        return self.__data.__iter__()


UNSET = object()


class Binder:
    """
    Idea stolen from https://martinfowler.com/eaaDev/uiArchs.html
    but I may have misunderstood.
    """

    def __init__(self, db, log, transforms=None, defaults=None):
        self._db = db
        self._log = log
        self.__data = {}
        self._widget_names = {}
        self.__defaults = defaults if defaults else {}
        self._clear_and_set_defaults()
        self._dbview = TransformedView(self.__data, transforms)

    def bind(self, widget, name, transforms=None, default=UNSET):
        """
        Call with each widget in turn.  The name should be the column name in
        the database.
        """
        if transforms is not None:
            self._dbview.set_transforms(name, transforms)
        if default is not UNSET:
            self.__defaults[name] = default
        self._log.debug('Binding %s as %s' % (widget, name))
        connect_signal(widget, 'change', self._save_widget_value)
        self._widget_names[widget] = name
        return widget

    def save(self, unused_widget):
        """
        Target for button click.
        """
        self._write_values_to_db()

    def reset(self, unused_Widget):
        """
        Target for button click.
        """
        self._read_values_from_db()

    @staticmethod
    def connect(widget, name, callback):
        """
        Shortcut for inline wiring of callbacks.
        """
        connect_signal(widget, name, callback)
        return widget

    def _changed(self, name, value):
        """
        Has the value for this name changed?
        """
        return name not in self.__data or value != self.__data[name]

    def _clear_and_set_defaults(self):
        """
        Called before read.
        """
        self.__data.clear()
        for name in self.__defaults:
            if name not in self.__data:
                self.__data[name] = self.__defaults[name]

    def _save_widget_value(self, widget, value):
        """
        Update the value in the store, but don't write to disk (although
        sub-classes may auto-save if the key changes).
        """
        self._log.debug('Saving %s=%s' % (self._widget_names[widget], value))
        self.__data[self._widget_names[widget]] = value

    def _broadcast_values_to_widgets(self):
        """
        Attempt to set values on registered widgets.  Called by subclasses
        after reading from disk.
        """
        for widget in self._widget_names:
            name = self._widget_names[widget]
            value = self.__data[name]
            self._log.debug('Setting %s=%s on %s' % (name, value, widget))
            if hasattr(widget, 'state'):
                widget.state = value
            elif hasattr(widget, 'set_edit_text'):
                widget.set_edit_text(value)
            else:
                self._log.error('Cannot set value on %s (%s)' % (widget, dir(widget)))

    def _write_values_to_db(self):
        """
        Subclasses must write data to database.
        """
        raise NotImplemented()

    def _read_values_from_db(self):
        """
        Subclasses must read data from the database.
        """
        raise NotImplemented()


class DynamicBinder(Binder):
    """
    A binder that updates when a key value changes.
    """

    def __init__(self, db, log, transforms=None, defaults=None):
        super().__init__(db, log, transforms=transforms, defaults=defaults)
        self._key_name = None
        self._key_widget = None

    def bind_key(self, widget, name):
        self._key_widget = widget
        self._key_name = name
        return self.bind(widget, name)

    def bootstrap(self, state):
        self._save_widget_value(self._key_widget, state)

    def _save_widget_value(self, widget, value):
        # only trigger database access if the key is changing
        if widget == self._key_widget and self._changed(self._key_name, value):
            if self._key_name in self._dbview:
                self._write_values_to_db()
            else:
                self._log.debug('Cannot save values as no previous value for key %s' % self._key_name)
            super()._save_widget_value(widget, value)
            self._read_values_from_db()
            self._broadcast_values_to_widgets()
        else:
            super()._save_widget_value(widget, value)


class SingleTableDynamic(DynamicBinder):
    """
    Read/write all values from a single table in the database.
    """

    def __init__(self, db, log, table, transforms=None, defaults=None):
        super().__init__(db, log, transforms=transforms, defaults=defaults)
        self._table = table

    def _write_values_to_db(self):
        cmd = 'replace into %s (' % self._table
        cmd += ', '.join(self._dbview.keys())
        cmd += ') values ('
        cmd += ', '.join(['?'] * (len(self._dbview)))
        cmd += ')'
        values = list(self._dbview.values())
        self._log.debug('%s / %s' % (cmd, values))
        self._db.execute(cmd, values)

    def _read_values_from_db(self):
        cmd = 'select '
        cmd += ', '.join(self._widget_names.values())
        cmd += ' from %s where %s = ?' % (self._table, self._key_name)
        values = [self._dbview[self._key_name]]
        self._log.debug('%s / %s' % (cmd, values))
        row = self._db.execute(cmd, values).fetchone()
        self._clear_and_set_defaults()
        if row:
            for name in self._widget_names.values():
                self._log.debug('Read %s=%s' % (name, row[name]))
                self._dbview[name] = row[name]
        # in case not include in read, or defaults used
        self._dbview[self._key_name] = values[0]


class SingleTableStatic(Binder):
    """
    A binder associated with a single table whose key does not change.
    Typically the key value(s) are provided via defaults because they're
    not mutable via widgets.
    """

    def __init__(self, db, log, table, key_names='id', transforms=None, defaults=None,
                 insert_callback=None, autosave=False):
        super().__init__(db, log, transforms=transforms, defaults=defaults)
        self._table = table
        self._key_names = (key_names,) if isinstance(key_names, str) else key_names
        self._insert_callback = insert_callback
        self._autosave = autosave

    def _have_all_keys(self):
        return reduce(and_, (name in self._dbview for name in self._key_names))

    def _save_widget_value(self, widget, value):
        super()._save_widget_value(widget, value)
        if self._autosave:
            if self._have_all_keys():
                self._write_values_to_db()
            else:
                self._log.warning('Not saving because missing key values (%s)' % list(self._dbview.keys()))

    def read_row(self, row):
        self._clear_and_set_defaults()
        if row:
            for name in row.keys():
                self._log.debug('Read %s=%s' % (name, row[name]))
                self._dbview[name] = row[name]
        self._broadcast_values_to_widgets()

    def _write_values_to_db(self):
        if self._dbview:
            replace = self._have_all_keys()
            cmd = '%s into %s (' % ('replace' if replace else 'insert', self._table)
            cmd += ', '.join(self._dbview.keys())
            cmd += ') values ('
            cmd += ', '.join(['?'] * (len(self._dbview)))
            cmd += ')'
            values = list(self._dbview.values())
            self._log.debug('%s / %s' % (cmd, values))
            self._db.execute(cmd, values)
            if not replace and len(self._key_names) == 1:
                self._dbview[self._key_names[0]] = self._db.execute('select last_insert_rowid()', ()).fetchone()[0]
                if self._insert_callback: self._insert_callback()

    def _read_values_from_db(self):
        if self._have_all_keys():
            cmd = 'select '
            names = list(self._key_names)
            names.extend(self._widget_names.values())
            cmd += ', '.join(names)
            cmd += ' from %s where ' % self._table
            for i, name in enumerate(self._key_names):
                cmd += 'and ' if i else ''
                cmd += '%s = ? ' % name
            values = [self._dbview[name] for name in self._key_names]
            self._log.debug('%s / %s' % (cmd, values))
            row = self._db.execute(cmd, values).fetchone()
        else:
            row = None
        self.read_row(row)




from json import dumps

from . import name_constant, add_enum_constant, set_constant
from ..stoats.calculate.climb import Climb, MAX_CLIMB_GRADIENT, MIN_CLIMB_GRADIENT, MAX_CLIMB_REVERSAL, \
    MIN_CLIMB_ELEVATION

CLIMB_CNAME = 'Climb'


def add_climb(s, activity_group, phi=0.7):
    '''
    Add the constants necessary to auto-detect climbs.
    '''
    activity_group_constraint = str(activity_group)
    climb_name = name_constant(CLIMB_CNAME, activity_group)
    climb = add_enum_constant(s, climb_name, Climb, single=True, constraint=activity_group_constraint,
                              description='Data needed to detect climbs - see Climb enum')
    set_constant(s, climb, dumps({'phi': phi,
                                  'min_gradient': MIN_CLIMB_GRADIENT,
                                  'max_gradient': MAX_CLIMB_GRADIENT,
                                  'min_elevation': MIN_CLIMB_ELEVATION,
                                  'max_reversal': MAX_CLIMB_REVERSAL}))




from ch2.stoats.names import DISTANCE, SPEED, ELEVATION, HR_IMPULSE_10, CADENCE, FITNESS, FATIGUE, ACTIVE_DISTANCE, \
    ACTIVE_TIME, POWER, W

# additional names not used within the database, but used for display by the notetbook
# (bokeh is easiest if dataframe columns are also labels).

WINDOW = '60s'
POW_MINUS_ONE = '\u207b\u00b9'
POW_TWO = '\u00b2'
MIN_PERIODS = 1

DISTANCE_KM = '%s / km' % DISTANCE
SPEED_KMH = '%s / kmh%s' % (SPEED, POW_MINUS_ONE)
MED_SPEED_KMH = 'M(%s) %s / kmh%s' % (WINDOW, SPEED, POW_MINUS_ONE)
ELEVATION_M = '%s / m' % ELEVATION
CLIMB_MS = 'Climb / ms%s' % POW_MINUS_ONE
MED_HR_IMPULSE_10 = 'M(%s) %s' % (WINDOW, HR_IMPULSE_10)
MED_CADENCE = 'M(%s) %s' % (WINDOW, CADENCE)
MED_POWER_W = 'M(%s) %s / %s' % (WINDOW, POWER, W)
LOG_FITNESS = 'Log %s' % FITNESS
LOG_FATIGUE = 'Log %s' % FATIGUE
ACTIVE_DISTANCE_KM = '%s / km' % ACTIVE_DISTANCE
ACTIVE_TIME_H = '%s / h' % ACTIVE_TIME




from abc import abstractmethod

from .. import DbPipeline
from ...fit.format.read import filtered_records
from ...fit.profile.profile import read_fit
from ...lib.date import to_time
from ...lib.io import for_modified_files


class AbortImport(Exception):
    pass


class AbortImportButMarkScanned(AbortImport):
    pass


class FileImporter(DbPipeline):
    '''
    Base class for importing from a files that have been modified.
    '''

    def _import_all(self, paths, force=False):
        with self._db.session_context() as s:
            for_modified_files(self._log, s, paths, self._callback(), self, force=force)

    def _callback(self):
        def callback(file):
            self._log.debug('Scanning %s' % file)
            with self._db.session_context() as s:
                try:
                    self._import_path(s, file)
                    return True
                except AbortImport as e:
                    self._log.debug('Aborted %s' % file)
                    return isinstance(e, AbortImportButMarkScanned)
        return callback

    @abstractmethod
    def _import_path(self, s, path):
        pass


class FitFileImporter(FileImporter):
    '''
    Extend FileImporter with utility methods related to FIT files.
    '''

    def _load_fit_file(self, path, *options):
        types, messages, records = filtered_records(self._log, read_fit(self._log, path))
        return [record.as_dict(*options)
                for _, _, record in sorted(records,
                                           key=lambda r: r[2].timestamp if r[2].timestamp else to_time(0.0))]

    def _first(self, path, records, *names):
        return self.__assert_contained(path, records, names, 0)

    def _last(self, path, records, *names):
        return self.__assert_contained(path, records, names, -1)

    def __assert_contained(self, path, records, names, index):
        try:
            return [record for record in records if record.name in names][index]
        except IndexError:
            self._log.debug(f'No {names} entry(s) in {path}')
            raise AbortImportButMarkScanned()




from contextlib import contextmanager

from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.functions import count

from . import *
from .support import Base
from ..command.args import DATABASE, NamespaceWithVariables, NO_OP, parser
from ..lib.log import make_log

# mention these so they are "created"
Source,  Interval,
ActivityGroup, ActivityJournal, ActivityTimespan,
Topic, TopicJournal,
StatisticName, StatisticJournal, StatisticJournalInteger, StatisticJournalFloat, StatisticJournalText,
Segment, SegmentJournal
Pipeline
MonitorJournal
Constant, SystemConstant
ActivitySimilarity, ActivityNearby


# https://stackoverflow.com/questions/13712381/how-to-turn-on-pragma-foreign-keys-on-in-sqlalchemy-migration-script-or-conf
@event.listens_for(Engine, "connect")
def fk_pragma_on_connect(dbapi_con, _con_record):
    cursor = dbapi_con.cursor()
    cursor.execute("PRAGMA foreign_keys=ON;")  # https://www.sqlite.org/pragma.html#pragma_foreign_keys
    cursor.execute("PRAGMA temp_store=MEMORY;")  # https://www.sqlite.org/pragma.html#pragma_temp_store
    cursor.execute("PRAGMA threads=4;")  # https://www.sqlite.org/pragma.html#pragma_threads
    cursor.execute("PRAGMA cache_size=-1000000;")  # 1GB  https://www.sqlite.org/pragma.html#pragma_cache_size
    cursor.execute("PRAGMA secure_delete=OFF;")  # https://www.sqlite.org/pragma.html#pragma_secure_delete
    # cursor.execute("PRAGMA journal_mode=WAL;")  # https://www.sqlite.org/wal.html
    cursor.close()


@event.listens_for(Engine, 'close')
def analyxe_pragma_on_close(dbapi_con, _con_record):
    cursor = dbapi_con.cursor()
    cursor.execute("PRAGMA optimize;")  # https://www.sqlite.org/pragma.html#pragma_optimize
    cursor.close()


class Database:

    def __init__(self, args, log):
        self._log = log
        path = args.file(DATABASE)
        self._log.info('Using database at %s' % path)
        self.engine = create_engine('sqlite:///%s' % path, echo=False)
        self.session = sessionmaker(bind=self.engine)
        self.__create_tables()

    def __create_tables(self):
        if self.is_empty(tables=True):
            self._log.info('Creating tables')
            Base.metadata.create_all(self.engine)

    @contextmanager
    def session_context(self):
        session = self.session()
        try:
            yield session
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def is_empty(self, tables=False):
        if tables:
            # https://stackoverflow.com/questions/33053241/sqlalchemy-if-table-does-not-exist
            return not self.engine.dialect.has_table(self.engine, Source.__tablename__)
        else:
            with self.session_context() as s:
                n_topics = s.query(count(Topic.id)).scalar()
                n_activities = s.query(count(ActivityGroup.id)).scalar()
                n_statistics = s.query(count(StatisticName.id)).scalar()
            return not (n_topics + n_activities + n_statistics)


def add(s, instance):
    s.add(instance)
    return instance


def connect(args):
    '''
    Bootstrap from commandline-like args.
    '''
    p = parser()
    if len(args) == 1:
        args = args[0].split()
    elif args:
        args = list(args)
    else:
        args = []
    args.append(NO_OP)
    ns = NamespaceWithVariables(p.parse_args(args))
    log = make_log(ns)
    db = Database(ns, log)
    return ns, log, db





from os.path import splitext, basename

from pygeotile.point import Point

from ..load import StatisticJournalLoader
from ..names import LATITUDE, LONGITUDE, M, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, ELEVATION, RAW_ELEVATION
from ..read import AbortImport, FitFileImporter
from ...fit.format.records import fix_degrees, merge_duplicates, no_bad_values
from ...lib.date import to_time
from ...sortem.bilinear import bilinear_elevation_from_constant
from ...squeal.database import add, Timestamp
from ...squeal.tables.activity import ActivityGroup, ActivityJournal, ActivityTimespan
from ...squeal.tables.source import Interval
from ...squeal.tables.statistic import StatisticJournalFloat, STATISTIC_JOURNAL_CLASSES


class ActivityImporter(FitFileImporter):

    def _on_init(self, *args, **kargs):
        super()._on_init(*args, **kargs)
        with self._db.session_context() as s:
            self.__oracle = bilinear_elevation_from_constant(self._log, s)

    def run(self, paths, force=False):
        if 'sport_to_activity' not in self._kargs:
            raise Exception('No map from sport to activity')
        self._import_all(paths, force=force)

    def _import_path(self, s, path):

        self._log.info('Importing activity data from %s' % path)

        sport_to_activity, record_to_db, add_elevation = self._parse_pipeline_arguments()
        records = self._load_fit_file(path, merge_duplicates, fix_degrees, no_bad_values)
        ajournal, activity_group, first_timestamp = self._create_activity(s, path, records, sport_to_activity)
        loader = StatisticJournalLoader(self._log, s, ActivityImporter)
        timespan, warned, last_timestamp = None, 0, to_time(0.0)

        for record in records:
            if record.name == 'event' or (record.name == 'record' and record.timestamp > last_timestamp):
                if record.name == 'event' and record.value.event == 'timer' and record.value.event_type == 'start':
                    if timespan:
                        self._log.warning('Ignoring start with no corresponding stop (possible lost data?)')
                    else:
                        timespan = add(s, ActivityTimespan(activity_journal=ajournal,
                                                           start=record.value.timestamp,
                                                           finish=record.value.timestamp))
                if record.name == 'record':
                    lat, lon, timestamp = None, None, record.value.timestamp
                    # customizable loader
                    for field, name, units, type in record_to_db:
                        value = record.data.get(field, None)
                        if value is not None:
                            value = value[0][0]
                            loader.add(name, units, None, activity_group, ajournal, value, timestamp, type)
                            if name == LATITUDE:
                                lat = value
                            elif name == LONGITUDE:
                                lon = value
                    # values derived from lat/lon
                    if lat is not None and lon is not None:
                        x, y = Point.from_latitude_longitude(lat, lon).meters
                        loader.add(SPHERICAL_MERCATOR_X, M, None, activity_group, ajournal, x, timestamp,
                                   StatisticJournalFloat)
                        loader.add(SPHERICAL_MERCATOR_Y, M, None, activity_group, ajournal, y, timestamp,
                                   StatisticJournalFloat)
                        if add_elevation:
                            elevation = self.__oracle.elevation(lat, lon)
                            if elevation:
                                loader.add(RAW_ELEVATION, M, None, activity_group, ajournal, elevation, timestamp,
                                           StatisticJournalFloat)
                if record.name == 'event' and record.value.event == 'timer' \
                        and record.value.event_type == 'stop_all':
                    if timespan:
                        timespan.finish = record.value.timestamp
                        ajournal.finish = record.value.timestamp
                        timespan = None
                    else:
                        self._log.debug('Ignoring stop with no corresponding start (possible lost data?)')
                if record.name == 'record':
                    last_timestamp = record.timestamp
            else:
                if record.name == 'record':
                    self._log.warning('Ignoring duplicate record data for %s at %s - some data may be missing' %
                                      (path, record.timestamp))

        loader.load()

        # manually clean out intervals because we're doing a fast load
        Interval.clean_times(self._log, s, first_timestamp, last_timestamp)

        # used by nearby calculations to avoid work
        # no need for constraint because ajournal is per-group
        # use this class so import itself is always clearly understood, even if the subclass changes.
        Timestamp.set(s, ActivityImporter, key=ajournal.id)

        # used by subclasses
        return ajournal, loader

    def _parse_pipeline_arguments(self):
        sport_to_activity = self._assert_karg('sport_to_activity')
        record_to_db = [(field, name, units, STATISTIC_JOURNAL_CLASSES[type])
                        for field, (name, units, type) in self._assert_karg('record_to_db').items()]
        add_elevation = not any(name == ELEVATION for (field, name, units, type) in record_to_db)
        return sport_to_activity, record_to_db, add_elevation

    def _create_activity(self, s, path, records, sport_to_activity):
        first_timestamp = self._first(path, records, 'event', 'record').value.timestamp
        sport = self._first(path, records, 'sport').value.sport.lower()
        activity_group = self._activity_group(s, path, sport, sport_to_activity)
        self._delete_journals(s, activity_group, first_timestamp)
        ajournal = add(s, ActivityJournal(activity_group=activity_group,
                                          start=first_timestamp, finish=first_timestamp,  # will be over-written later
                                          fit_file=path, name=splitext(basename(path))[0]))
        return ajournal, activity_group, first_timestamp

    def _activity_group(self, s, path, sport, sport_to_activity):
        if sport in sport_to_activity:
            return self._lookup_activity_group(s, sport_to_activity[sport])
        else:
            self._log.warning('Unrecognised sport: "%s" in %s' % (sport, path))
            raise AbortImport()

    def _lookup_activity_group(self, s, name):
        activity_group = s.query(ActivityGroup).filter(ActivityGroup.name == name).one_or_none()
        if not activity_group:
            activities = s.query(ActivityGroup).all()
            if activities:
                self._log.info('Available activity group:')
                for activity_group in activities:
                    self._log.info('%s - %s' % (activity_group.name, activity_group.description))
            else:
                self._log.error('No activity groups defined - configure system correctly')
            raise Exception('ActivityGroup "%s" is not defined' % name)
        return activity_group

    def _delete_journals(self, s, activity_group, first_timestamp):
        # need to iterate because sqlite doesn't support multi-table delete and we have inheritance.
        for journal in s.query(ActivityJournal). \
                filter(ActivityJournal.activity_group == activity_group,
                       ActivityJournal.start == first_timestamp).all():
            Timestamp.clear(s, owner=ActivityImporter, key=journal.id)
            s.delete(journal)
        s.flush()




from collections import namedtuple

# a climb of 80m is roughly equivalent to a score of 8000 on strava's weird approach -
# https://support.strava.com/hc/en-us/articles/216917057-How-are-Strava-climbs-categorized-For-Rides-
MIN_CLIMB_ELEVATION = 80
MIN_CLIMB_GRADIENT = 3
MAX_CLIMB_GRADIENT = 40
MAX_CLIMB_REVERSAL = 0.1

# trade-off between pure elevation (0) and pure gradient (1)
CLIMB_PHI = 0.6


Climb = namedtuple('Climb', 'phi, min_elevation, min_gradient, max_gradient, max_reversal',
                   defaults=(CLIMB_PHI, MIN_CLIMB_ELEVATION, MIN_CLIMB_GRADIENT,
                             MAX_CLIMB_GRADIENT, MAX_CLIMB_REVERSAL))


def find_climbs(waypoints, params=Climb()):
    waypoints = [w for w in waypoints if w.elevation is not None]
    if waypoints:
        mn, mx = min(w.elevation for w in waypoints), max(w.elevation for w in waypoints)
        if mx - mn > params.min_elevation:
            score, lo, hi = biggest_climb(waypoints, params=params)
            if score:
                a, b, c = split(waypoints, lo, hi)
                yield from find_climbs(a, params=params)
                yield from contiguous(b, params=params)
                yield from find_climbs(c, params=params)


def split(waypoints, lo, hi, inside=True):
    i = waypoints.index(lo)
    j = waypoints.index(hi)
    if i > j:
        i, j = j, i
    if inside:
        j += 1
    else:
        i += 1
    return waypoints[:i], waypoints[i:j], waypoints[j:]


def contiguous(waypoints, params=Climb()):
    up = waypoints[-1].elevation - waypoints[0].elevation
    if up >= params.min_elevation:
        down, lo, hi = biggest_reversal(waypoints)
        if down and down > params.max_reversal * up:
            a, b, c = split(waypoints, lo, hi, inside=False)
            yield from contiguous(a, params=params)
            yield from find_climbs(b, params=params)
            yield from contiguous(c, params=params)
        else:
            along = waypoints[-1].distance - waypoints[0].distance
            if along and 100 * up / along < params.max_gradient:
                yield waypoints[0], waypoints[-1]


def sort(waypoints, reverse=False):
    return sorted(waypoints, key=lambda w: w.elevation, reverse=reverse)


def first_or_none(generator):
    try:
        return next(generator)
    except StopIteration:
        return None


def biggest_reversal(waypoints):
    best = None, None, None
    if waypoints:
        highest = sort(waypoints, reverse=True)
        lowest = sort(waypoints)
        for hi in highest:
            lo = first_or_none(l for l in lowest if l.distance > hi.distance and l.elevation < hi.elevation)
            if lo:
                drop = hi.elevation - lo.elevation
                # if not reverse: print(score, hi.elevation, lo.elevation)
                if best[0] is None or drop > best[0]:
                    best = (drop, lo, hi)
            if best[0] and best[0] > hi.elevation - lowest[0].elevation:
                break  # exit if there is no way to improve
    return best


def biggest_climb(waypoints, params=Climb(), grid=10):
    if len(waypoints) > 100 * grid:
        found, lo, hi = search(waypoints[::grid], params=params)
        if found:
            i, j = waypoints.index(lo), waypoints.index(hi)
            if i + grid >= j - grid:
                waypoints = waypoints[max(i-grid, 0):j+grid]
            else:
                waypoints = waypoints[max(i-grid, 0):i+grid] + waypoints[max(j-grid, 0):j+grid]
        else:
            return None, None, None
    return search(waypoints, params=params)


def search(waypoints, params=Climb()):
    best = None, None, None
    if waypoints:
        highest = sort(waypoints, reverse=True)
        lowest = sort(waypoints)
        for hi in highest:
            # use distance rather than time to avoid division by zero with limited resolution distance
            for lo in filter(lambda lo: lo.distance < hi.distance and
                                        hi.elevation - lo.elevation > params.min_elevation and
                                        100 * (hi.elevation - lo.elevation) / (hi.distance - lo.distance) >
                                        params.min_gradient,
                             lowest):
                score = (hi.elevation - lo.elevation) / (hi.distance - lo.distance) ** params.phi
                if best[0] is None or score > best[0]:
                    best = (score, lo, hi)
            if (hi.elevation - lowest[0].elevation) < params.min_elevation:
                break  # abort if there is no valid future value
    return best




from tempfile import NamedTemporaryFile
from unittest import TestCase

from ch2.command.args import bootstrap_file, m, V, mm, DEV
from ch2.config.default import default
from ch2.config.plan.british import twelve_week_improver
from ch2.config.plan.exponential import exponential_time, exponential_distance
from ch2.lib.date import to_date, add_date
from ch2.squeal.tables.topic import Topic


class TestPlan(TestCase):

    def test_british(self):
        with NamedTemporaryFile() as f:
            args, log, db = bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)
            plan = twelve_week_improver('2018-07-25')
            plan.create(log, db)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with db.session_context() as s:
                root = s.query(Topic).filter(Topic.parent_id == None, Topic.name == 'Plan').one()
                self.assertEqual(len(root.children), 1)
                self.assertTrue(root.schedule)
                self.assertEqual(root.schedule.start, to_date('2018-07-25'))
                self.assertEqual(root.schedule.finish, add_date('2018-07-25', (12, 'w')))
                parent = root.children[0]
                self.assertEqual(len(parent.children), 7)
                for child in parent.children:
                    print(child)

    def test_exponential_time(self):
        with NamedTemporaryFile() as f:
            args, log, db = bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)
            plan = exponential_time('Time test', '2d[2]', '20M', 5, '2018-07-25', '3m')
            plan.create(log, db)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with db.session_context() as s:
                root = s.query(Topic).filter(Topic.parent_id == None, Topic.name == 'Plan').one()
                self.assertEqual(len(root.children), 1)
                parent = root.children[0]
                self.assertEqual(len(parent.children), 46)
                for child in parent.children:
                    print(child)

    def test_exponential_distance(self):
        with NamedTemporaryFile() as f:
            args, log, db = bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)
            plan = exponential_distance('Distance test', 'w[mon,wed,fri]', '20km', 5, '2018-07-25', '1m')
            plan.create(log, db)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with db.session_context() as s:
                root = s.query(Topic).filter(Topic.parent_id == None, Topic.name == 'Plan').one()
                self.assertEqual(len(root.children), 1)
                parent = root.children[0]
                self.assertEqual(len(parent.children), 14)
                for child in parent.children:
                    print(child)




from abc import abstractmethod

from sqlalchemy import not_
from sqlalchemy.sql.functions import count

from ch2.stoats.load import StatisticJournalLoader
from .. import DbPipeline
from ..waypoint import WaypointReader
from ...lib.schedule import Schedule
from ...squeal import ActivityJournal, ActivityGroup, Pipeline, Interval, Timestamp, StatisticJournal, \
    StatisticName, StatisticJournalFloat, StatisticJournalInteger
from ...squeal.types import short_cls


def run_pipeline_after(log, db, type, after=None, force=False, like=None):
    with db.session_context() as s:
        for cls, args, kargs in Pipeline.all(log, s, type, like=like):
            log.info('Running %s (%s, %s)' % (short_cls(cls), args, kargs))
            cls(log, db, *args, **kargs).run(force=force, after=after)


def run_pipeline_paths(log, db, type, paths, force=False, like=None):
    with db.session_context() as s:
        for cls, args, kargs in Pipeline.all(log, s, type, like=like):
            log.info('Running %s (%s, %s)' % (short_cls(cls), args, kargs))
            cls(log, db, *args, **kargs).run(paths, force=force)


class IntervalCalculator(DbPipeline):
    '''
    Support for calculations associated with intervals.
    '''

    def run(self, force=False, after=None):
        schedule = Schedule(self._assert_karg('schedule'))
        self.run_schedule(force=force, after=after, schedule=schedule)

    def run_schedule(self, force=False, after=None, schedule=None):
        if force:
            self._delete(after=after)
        self._run_calculations(schedule)

    @abstractmethod
    def _run_calculations(self, schedule):
        raise NotImplementedError()

    def _delete(self, after=None):
        self._delete_intervals(after)

    def _delete_intervals(self, after=None):
        # we delete the intervals that all summary statistics depend on and they will cascade
        with self._db.session_context() as s:
            for repeat in range(2):
                if repeat:
                    q = s.query(Interval)
                else:
                    q = s.query(count(Interval.id))
                q = self._filter_intervals(q)
                if after:
                    q = q.filter(Interval.finish > after)
                if repeat:
                    for interval in q.all():
                        self._log.debug('Deleting %s' % interval)
                        s.delete(interval)
                else:
                    n = q.scalar()
                    if n:
                        self._log.warning('Deleting %d intervals' % n)
                    else:
                        self._log.warning('No intervals to delete')

    def _filter_intervals(self, q):
        return q.filter(Interval.owner == self)


class ActivityCalculator(DbPipeline):
    '''
    Support for calculations associated with activity journals (which is most).
    '''

    def run(self, force=False, after=None):
        with self._db.session_context() as s:
            for activity_group in s.query(ActivityGroup).all():
                self._log.debug('Checking statistics for activity group %s' % activity_group.name)
                if force:
                    self._delete_my_statistics(s, activity_group, after=after)
                self._run_activity(s, activity_group)

    def _run_activity(self, s, activity_group):
        for ajournal in self._activity_journals_with_missing_data(s, activity_group):
            # set constraint so we can delete on force
            with Timestamp(owner=self, constraint=activity_group, key=ajournal.id).on_success(s):
                self._log.info('Running %s for %s' % (short_cls(self), ajournal))
                self._add_stats(s, ajournal)

    def _activity_journals_with_missing_data(self, s, activity_group):
        existing_ids = s.query(Timestamp.key). \
            filter(Timestamp.owner == self,
                   Timestamp.constraint == activity_group).cte()
        yield from s.query(ActivityJournal). \
            filter(not_(ActivityJournal.id.in_(existing_ids)),
                   ActivityJournal.activity_group == activity_group). \
            order_by(ActivityJournal.start).all()

    @abstractmethod
    def _filter_statistic_journals(self, q):
        raise NotImplementedError()

    @abstractmethod
    def _add_stats(self, s, ajournal):
        raise NotImplementedError()

    def _delete_my_statistics(self, s, agroup, after=None):
        '''
        Delete all statistics owned by this class and in the activity group.
        Fast because in-SQL.
        '''
        s.commit()   # so that we don't have any risk of having something in the session that can be deleted
        for repeat in range(2):
            cte = s.query(StatisticName.id).filter(StatisticName.owner == self)
            if repeat:
                q = s.query(StatisticJournal)
            else:
                q = s.query(count(StatisticJournal.id))
            q = q.filter(StatisticJournal.statistic_name_id.in_(cte.cte()))
            q = self._constrain_source(s, q, agroup)
            if after:
                q = q.filter(StatisticJournal.time >= after)
            if repeat:
                q.delete(synchronize_session=False)
            else:
                n = q.scalar()
                if n:
                    self._log.warning('Deleting %d statistics for %s' % (n, agroup))
                else:
                    self._log.warning('No statistics to delete for %s' % agroup)
        Timestamp.clear_after(s, after, self, constraint=agroup)
        s.commit()

    def _constrain_source(self, s, q, agroup):
        cte = s.query(ActivityJournal.id).filter(ActivityJournal.activity_group_id == agroup.id).cte()
        return q.filter(StatisticJournal.source_id.in_(cte))


class WaypointCalculator(ActivityCalculator):
    '''
    Original calculator scheme, still used by most code.  Pure-python and SQLAlchemy,
    '''

    def _add_stats(self, s, ajournal):
        owner = self._assert_karg('owner')
        waypoints = list(WaypointReader(self._log).read(s, ajournal, self._names(), owner))
        if waypoints:
            self._add_stats_from_waypoints(s, ajournal, waypoints)
        else:
            self._log.warning('No statistics for %s' % ajournal)

    @abstractmethod
    def _add_stats_from_waypoints(self, s, ajournal, waypoints):
        raise NotImplementedError()

    @abstractmethod
    def _names(self):
        raise NotImplementedError()


class DataFrameCalculator(ActivityCalculator):
    '''
    New calculator scheme.  Uses data frames / shares code with analysis.
    '''

    def _add_stats(self, s, ajournal):
        df = self._load_data(s, ajournal)
        if df is not None and len(df):
            df = self._extend_data(s, df)
            loader = StatisticJournalLoader(self._log, s, self)
            self._copy_results(s, ajournal, df, loader)
            loader.load()
        else:
            self._log.warning('No statistics for %s' % ajournal)

    @abstractmethod
    def _load_data(self, s, ajournal):
        raise NotImplementedError()

    @abstractmethod
    def _extend_data(self, s, df):
        raise NotImplementedError()

    @abstractmethod
    def _copy_results(self, s, ajournal, df, loader):
        raise NotImplementedError()




from itertools import groupby

from ..calculate.activity import ActivityStatistics
from ..names import CLIMB_ELEVATION, TOTAL_CLIMB
from ...squeal import StatisticJournal, StatisticName


def climbs_for_activity(s, ajournal):
    total = s.query(StatisticJournal).join(StatisticName). \
        filter(StatisticName.name == TOTAL_CLIMB,
               StatisticJournal.time == ajournal.start,
               StatisticName.owner == ActivityStatistics,
               StatisticName.constraint == ajournal.activity_group).order_by(StatisticJournal.time).one_or_none()
    statistics = s.query(StatisticJournal).join(StatisticName). \
        filter(StatisticName.name.like('Climb %'),
               StatisticJournal.time >= ajournal.start,
               StatisticJournal.time <= ajournal.finish,
               StatisticName.owner == ActivityStatistics,
               StatisticName.constraint == ajournal.activity_group).order_by(StatisticJournal.time).all()
    return total, sorted((dict((statistic.statistic_name.name, statistic) for statistic in climb_statistics)
                          for _, climb_statistics in groupby(statistics, key=lambda statistic: statistic.time)),
                         key=lambda climb: climb[CLIMB_ELEVATION].value, reverse=True)




from urwid import Text, Columns, Pile

from . import Displayer
from ..calculate.monitor import MonitorStatistics
from ..names import REST_HR, DAILY_STEPS
from ...squeal.tables.statistic import StatisticJournal, StatisticName
from ...uweird.fields import ReadOnlyField
from ...uweird.fields.summary import summary_columns
from ...uweird.tui.decorators import Indent


class MonitorDiary(Displayer):

    def _display_date(self, s, f, date):
        columns = self.__fields(s, date)
        if columns:
            yield Pile([Text('Monitor'),
                        Indent(Columns(columns))])

    def __fields(self, s, date):
        steps = self.__field(s, date, DAILY_STEPS)
        rest_hr = self.__field(s, date, REST_HR)
        if steps or rest_hr:
            return [steps if steps else Text(''), rest_hr if rest_hr else Text('')]
        else:
            return None

    def __field(self, s, date, name):
        sjournal = StatisticJournal.at_date(s, date, name, MonitorStatistics, None)
        if sjournal:
            return ReadOnlyField(self._log, sjournal, date=date).widget()
        else:
            return None

    def _display_schedule(self, s, f, date, schedule=None):
        columns = list(self.__schedule_fields(s, f, date, schedule))
        if columns:
            yield Pile([Text('Monitor'),
                        Indent(Pile(columns))])

    def __schedule_fields(self, s, f, date, schedule):
        names = list(self.__names(s, DAILY_STEPS, REST_HR))
        yield from summary_columns(self._log, s, f, date, schedule, names)

    def __names(self, s, *names):
        for name in names:
            sname = s.query(StatisticName). \
                filter(StatisticName.name == name,
                       StatisticName.owner == MonitorStatistics).one_or_none()
            if sname:
                yield sname




import datetime as dt
from collections import defaultdict

from cachetools import cached
from sqlalchemy import desc

from ..names import HEART_RATE, BPM, STEPS, STEPS_UNITS, ACTIVITY, CUMULATIVE_STEPS_START, \
    CUMULATIVE_STEPS_FINISH
from ..read import AbortImportButMarkScanned, AbortImport, FitFileImporter
from ...fit.format.records import fix_degrees, unpack_single_bytes, merge_duplicates
from ...lib.date import time_to_local_date, format_time
from ...squeal.database import add, Timestamp
from ...squeal.tables.monitor import MonitorJournal
from ...squeal.tables.statistic import StatisticJournalInteger, StatisticJournalText, StatisticName, StatisticJournal

ACTIVITY_TYPE_ATTR = 'activity_type'
HEART_RATE_ATTR = 'heart_rate'
MONITORING_ATTR = 'monitoring'
MONITORING_INFO_ATTR = 'monitoring_info'
STEPS_ATTR = 'steps'


class MonitorImporter(FitFileImporter):

    # the monitor steps data are cumulative, but we want incremental.
    # that's easy to do within a single file, but to be correct across files we also
    # store the cumulative value at the start and end,  we use these to "patch things up"
    # is we read a missing file.

    def _on_init(self, *args, **kargs):
        super()._on_init(*args, **kargs)

        @cached(cache={}, key=lambda s, name, units, summary, constraint: (name, constraint))
        def statistics_cache(s, name, units, summary, constraint):
            return StatisticName.add_if_missing(self._log, s, name, units, summary, self, constraint)

        self.__statistics_cache = statistics_cache

    def run(self, paths, force=False):
        self._import_all(paths, force=force)

    def _create(self, s, name, units, summary, constraint, source, value, time, type):
        return type(statistic_name=self.__statistics_cache(s, name, units, summary, constraint),
                    source=source, value=value, time=time)

    def _add(self, s, name, units, summary, constraint, source, value, time, type):
        return add(s, self._create(s, name, units, summary, constraint, source, value, time, type))

    def _check_contains(self, s, start, finish, path):
        for mjournal in s.query(MonitorJournal). \
                filter(MonitorJournal.start >= start,
                       MonitorJournal.finish <= finish).all():
            self._log.warning('Replacing %s with data from %s for %s - %s' % (mjournal, path, start, finish))
            s.delete(mjournal)

    def _check_inside(self, s, start, finish, path):
        for mjournal in s.query(MonitorJournal). \
                filter(MonitorJournal.start <= start,
                       MonitorJournal.finish >= finish).all():
            self._log.warning('%s already includes data from %s for %s - %s' % (mjournal, path, start, finish))
            raise AbortImportButMarkScanned()

    def _check_overlap(self, s, start, finish, path):
        for mjournal in s.query(MonitorJournal). \
                filter(MonitorJournal.start < finish,
                       MonitorJournal.finish > start).all():
            self._log.warning('%s overlaps data from %s for %s - %s' % (mjournal, path, start, finish))
            self._log.error('Conflict between %s and %s' % (path, mjournal.fit_file))
            raise AbortImport()

    def _check_previous(self, s, start, finish, path):
        self._check_contains(s, start, finish, path)
        self._check_inside(s, start, finish, path)
        self._check_overlap(s, start, finish, path)

    def _delete_journals(self, s, first_timestamp, path):
        # key only on time so that repeated files don't affect things
        if not first_timestamp:
            raise Exception('Missing timestamp in %s' % path)
        # need to iterate because sqlite doesn't support multi-table delete and we have inheritance.
        for mjournal in s.query(MonitorJournal). \
                filter(MonitorJournal.start == first_timestamp).all():
            self._log.debug('Deleting %s' % mjournal)
            s.delete(mjournal)
        s.flush()

    def _delta(self, steps, activity, cumulative):
        if steps >= cumulative[activity]:
            delta = steps - cumulative[activity]
        else:
            delta = steps
        cumulative[activity] = steps
        return delta

    def _save_cumulative(self, s, time, cumulative, mjournal, name):
        self._log.debug('Adding %s at time %s' % (name, format_time(time)))
        for activity in cumulative:
            self._add(s, name, STEPS_UNITS, None, activity,
                      mjournal, cumulative[activity], time, StatisticJournalInteger)
        self._log.debug('Added: %s' % ', '.join('%s:%s' % item for item in cumulative.items()))

    def _update_cumulative(self, s, time, cumulative, mjournal, name, saved):
        # go back and fill in with zeroes and missing starting values
        for activity in saved:
            del cumulative[activity]
        if cumulative:
            for activity in cumulative:
                cumulative[activity] = 0
            self._save_cumulative(s, time, cumulative, mjournal, name)

    def _read_cumulative(self, s, time, cumulative, name):
        # no constraint, so all activities
        for journal in s.query(StatisticJournalInteger).join(StatisticName). \
                filter(StatisticJournalInteger.time == time,
                       StatisticName.name == name,
                       StatisticName.owner == self).all():
            cumulative[journal.statistic_name.constraint] = journal.value
        self._log.debug('Read %s at %s' % (name, format_time(time)))
        self._log.debug('Read: %s' % ', '.join('%s:%s' % item for item in cumulative.items()))

    def _read_previous(self, s, first_timestamp, cumulative):
        prev = s.query(MonitorJournal). \
            filter(MonitorJournal.finish <= first_timestamp). \
            order_by(desc(MonitorJournal.finish)).limit(1).one_or_none()
        if prev:
            self._read_cumulative(s, prev.finish, cumulative, CUMULATIVE_STEPS_FINISH)

    def _update_next(self, s, last_timestamp, cumulative):
        self._log.debug('Checking for following file after %s' % format_time(last_timestamp))
        next = s.query(MonitorJournal). \
            filter(MonitorJournal.start >= last_timestamp). \
            order_by(MonitorJournal.start).limit(1).one_or_none()
        if next:
            self._log.debug('Found file at %s - fixing up' % format_time(next.start))
            # no constraint, so all activities
            for cumulative_journal in s.query(StatisticJournalInteger).join(StatisticName). \
                    filter(StatisticJournalInteger.time == next.start,
                           StatisticName.name == CUMULATIVE_STEPS_START,
                           StatisticName.owner == self).all():
                activity = cumulative_journal.statistic_name.constraint
                if activity in cumulative and cumulative[activity] != cumulative_journal.value:
                    self._log.debug('Found %s with inconsistent value (%s != %s)' %
                                    (CUMULATIVE_STEPS_START, cumulative_journal.value, cumulative[activity]))
                    steps_journal = StatisticJournal.after(s, next.start, STEPS, self, activity)
                    self._log.debug('Found %s with value %s' % (STEPS, steps_journal.value))
                    # only fix up if the cumulative value didn't reset
                    if steps_journal.value + cumulative_journal.value >= cumulative[activity]:
                        steps_journal.value = steps_journal.value - cumulative[activity] + cumulative_journal.value
                        self._log.debug('Updated %s at %s to %s' %
                                        (STEPS, format_time(last_timestamp), steps_journal.value))
                        cumulative_journal.value = cumulative[activity]
                        if steps_journal.value == 0:
                            activity_journal = StatisticJournal.at(s, steps_journal.time, ACTIVITY, self, activity)
                            s.delete(steps_journal)
                            s.delete(activity_journal)
                            self._log.debug('Deleted %s and %s' % (STEPS, ACTIVITY))

    def _create_journals(self, s, records, cumulative, mjournal):
        n_heart_rate, n_steps, steps_journals = 0, 0, defaultdict(lambda: {})
        for record in records:
            if HEART_RATE_ATTR in record.data:
                self._add(s, HEART_RATE, BPM, None, None, mjournal, record.data[HEART_RATE_ATTR][0][0],
                          record.timestamp, StatisticJournalInteger)
                n_heart_rate += 1
            if STEPS_ATTR in record.data:
                for (activity, steps) in zip(record.data[ACTIVITY_TYPE_ATTR][0], record.data[STEPS_ATTR][0]):
                    steps = self._delta(steps, activity, cumulative)
                    if steps:
                        if activity in steps_journals[record.timestamp]:
                            # sometimes get values at exactly the same time :(
                            steps_journals[record.timestamp][activity].value += steps
                        else:
                            steps_journals[record.timestamp][activity] = \
                                self._create(s, STEPS, STEPS_UNITS, None, activity,
                                             mjournal, steps, record.timestamp,
                                             StatisticJournalInteger)
                        n_steps += 1
        self._log.debug('Found %d steps and %d heart rate values' % (n_heart_rate, n_steps))
        return steps_journals

    def _merge_boundary(self, s, steps_journals, time):
        if time in steps_journals:
            for activity in list(steps_journals[time].keys()):
                sjournal = StatisticJournal.at(s, time, STEPS, self, activity)
                if sjournal:
                    if sjournal.value == steps_journals[time][activity].value:
                        self._log.debug('Matching data at %s' % format_time(time))
                    else:
                        # we have a contradiction, so simply use the latest
                        # could maybe use max() instead?
                        self._log.warning('Replacing %s data at %s (%s replaced by %s)' %
                                       (STEPS, format_time(time), sjournal.value,
                                        steps_journals[time][activity].value))
                        sjournal.value = steps_journals[time][activity].value
                    del steps_journals[time][activity]

    def _import_path(self, s, path):

        records = self._load_fit_file(path, merge_duplicates, fix_degrees, unpack_single_bytes)

        first_timestamp = self._first(path, records, MONITORING_INFO_ATTR).timestamp
        last_timestamp = self._last(path, records, MONITORING_ATTR).timestamp
        if first_timestamp == last_timestamp:
            self._log.debug('File %s is empty (no timespan)' % path)
            raise AbortImportButMarkScanned()

        self._log.info('Importing monitor data from %s for %s - %s' %
                       (path, format_time(first_timestamp), format_time(last_timestamp)))
        self._check_previous(s, first_timestamp, last_timestamp, path)
        mjournal = add(s, MonitorJournal(start=first_timestamp, fit_file=path, finish=last_timestamp))

        cumulative = defaultdict(lambda: 0)
        self._read_previous(s, first_timestamp, cumulative)
        self._save_cumulative(s, first_timestamp, cumulative, mjournal, CUMULATIVE_STEPS_START)
        saved_activities = list(cumulative.keys())

        steps_journals = self._create_journals(s, records, cumulative, mjournal)

        self._save_cumulative(s, last_timestamp, cumulative, mjournal, CUMULATIVE_STEPS_FINISH)
        self._update_next(s, last_timestamp, cumulative)
        self._update_cumulative(s, first_timestamp, cumulative, mjournal, CUMULATIVE_STEPS_START, saved_activities)

        self._merge_boundary(s, steps_journals, first_timestamp)
        self._merge_boundary(s, steps_journals, last_timestamp)

        # write steps at end, when adjoining data have been fixed
        self._log.debug('Adding %s to database' % STEPS)
        for timestamp in steps_journals:
            for activity in steps_journals[timestamp]:
                add(s, steps_journals[timestamp][activity])
                self._add(s, ACTIVITY, None, None, activity, mjournal,
                          activity, timestamp, StatisticJournalText)

        Timestamp.set(s, self, key=mjournal.id)


def missing_dates(log, s):
    # we don't try to be perfect here.  the idea is that it's called to get the latest
    # updates, rather than fill in all gaps (do the bulk download thing for that).
    # we also don't try to get fractional data.
    # and as for timezones... we just assume garmin uses the local timezone.
    latest = s.query(MonitorJournal).order_by(desc(MonitorJournal.start)).limit(1).one_or_none()
    if latest is None:
        log.warning('No existing monitor data - ' +
                 'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
        return
    # find the mid-point to avoid any problems with timezones and edge cases
    # (these files don't span more than a day)
    seconds = (latest.finish - latest.start).total_seconds() / 2
    start = time_to_local_date(latest.start + dt.timedelta(seconds=seconds) + dt.timedelta(days=1))
    finish = dt.date.today()
    days = (finish - start).days
    if days > 10:
        raise Exception('Too many days (%d) - ' % days +
                        'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
    if days:
        # exclude today since it will be incomplete
        while start < finish:
            yield start
            start += dt.timedelta(days=1)
    else:
        log.warning('No dates to download')




from subprocess import run
from tempfile import NamedTemporaryFile
from unittest import TestCase

from sqlalchemy.sql.functions import count

from ch2.command.args import m, V, bootstrap_file
from ch2.config.personal import acooke
from ch2.lib.date import to_time, local_date_to_time, to_date
from ch2.squeal.database import add
from ch2.squeal.tables.source import Source, Interval
from ch2.squeal.tables.statistic import StatisticJournalText, StatisticJournal, StatisticJournalFloat, StatisticName, \
    StatisticJournalInteger, StatisticJournalType
from ch2.squeal.tables.topic import TopicJournal, Topic
from ch2.stoats.calculate.summary import SummaryStatistics


# the idea here is to test the new database schema with sources etc
# so we configure a database then load some data, calculate some stats,
# and see if everything works as expected.

class TestSources(TestCase):

    def test_sources(self):

        with NamedTemporaryFile() as f:

            args, log, db = bootstrap_file(f, m(V), '5', configurator=acooke)

            with db.session_context() as s:

                # add a diary entry

                diary = s.query(Topic).filter(Topic.name == 'Diary').one()
                d = add(s, TopicJournal(topic=diary, date='2018-09-29'))
                d.populate(log, s)
                self.assertEqual(len(d.topic.fields), 9, list(enumerate(map(str, d.topic.fields))))
                self.assertEqual(d.topic.fields[0].statistic_name.name, 'Notes')
                self.assertEqual(d.topic.fields[1].statistic_name.name, 'Weight', str(d.topic.fields[1]))
                for field in d.topic.fields:
                    if field in d.statistics:
                        self.assertTrue(d.statistics[field].value is None, field)
                d.statistics[d.topic.fields[0]].value = 'hello world'
                d.statistics[d.topic.fields[1]].value = 64.5

            with db.session_context() as s:

                # check the diary entry was persisted

                diary = s.query(Topic).filter(Topic.name == 'Diary').one()
                d = s.query(TopicJournal).filter(TopicJournal.topic == diary,
                                                 TopicJournal.date == '2018-09-29').one()
                s.flush()
                d.populate(log, s)
                self.assertEqual(len(d.topic.fields), 9, list(enumerate(map(str, d.topic.fields))))
                self.assertEqual(d.topic.fields[0].statistic_name.name, 'Notes')
                self.assertEqual(d.statistics[d.topic.fields[0]].value, 'hello world')
                self.assertEqual(d.topic.fields[1].statistic_name.name, 'Weight')
                self.assertEqual(d.statistics[d.topic.fields[1]].value, 64.5)
                self.assertEqual(d.statistics[d.topic.fields[1]].type, StatisticJournalType.FLOAT)

            # generate summary stats

            SummaryStatistics(log, db, schedule='m').run()
            SummaryStatistics(log, db, schedule='y').run()

            with db.session_context() as s:

                # check the summary stats

                diary = s.query(Topic).filter(Topic.name == 'Diary').one()
                weight = s.query(StatisticJournal).join(StatisticName). \
                    filter(StatisticName.owner == diary, StatisticName.name == 'Weight').one()
                self.assertEqual(weight.value, 64.5)
                self.assertEqual(len(weight.measures), 2, weight.measures)
                self.assertEqual(weight.measures[0].rank, 1)
                self.assertEqual(weight.measures[0].percentile, 100, weight.measures[0].percentile)
                n = s.query(count(StatisticJournalFloat.id)).scalar()
                self.assertEqual(n, 4, n)
                n = s.query(count(StatisticJournalInteger.id)).scalar()
                self.assertEqual(n, 11, n)
                m_avg = s.query(StatisticJournalFloat).join(StatisticName). \
                    filter(StatisticName.name == 'Avg/Month Weight').one()
                self.assertEqual(m_avg.value, 64.5)
                y_avg = s.query(StatisticJournalFloat).join(StatisticName). \
                    filter(StatisticName.name == 'Avg/Year Weight').one()
                self.assertEqual(y_avg.value, 64.5)
                month = s.query(Interval).filter(Interval.schedule == 'm').one()
                self.assertEqual(month.start, to_date('2018-09-01'), month.start)
                self.assertEqual(month.finish, to_date('2018-10-01'), month.finish)

            with db.session_context() as s:

                # delete the diary entry

                diary = s.query(Topic).filter(Topic.name == 'Diary').one()
                d = s.query(TopicJournal).filter(TopicJournal.topic == diary,
                                                 TopicJournal.date == '2018-09-29').one()
                s.delete(d)

            run('sqlite3 %s ".dump"' % f.name, shell=True)

            with db.session_context() as s:

                # check the delete cascade

                self.assertEqual(s.query(count(TopicJournal.id)).scalar(), 0)
                # this should be zero because the Intervals were automatically deleted
                for source in s.query(Source).all():
                    print(source)
                for journal in s.query(StatisticJournal).all():
                    print(journal)
                self.assertEqual(s.query(count(Source.id)).scalar(), 8, list(map(str, s.query(Source).all())))  # constants
                self.assertEqual(s.query(count(StatisticJournalText.id)).scalar(), 5, s.query(count(StatisticJournalText.id)).scalar())
                self.assertEqual(s.query(count(StatisticJournal.id)).scalar(), 5, s.query(count(StatisticJournal.id)).scalar())




from collections import namedtuple
from json import loads

import numpy as np
import pandas as pd

from . import DataFrameCalculator
from ...data import activity_statistics, DISTANCE, ELEVATION, SPEED, POW_TWO, TIME, TIMESPAN_ID, AVG, W, POWER
from ...squeal import StatisticJournalFloat, Constant


def d(name): return f'Delta {name}'


def avg(name): return f'Avg {name}'


ENERGY = 'Energy'
LOSS = 'Loss'
SPEED_2 = f'{SPEED}{POW_TWO}'
CDA = 'CdA'
CRR = 'Crr'

AVG_SPEED_2 = avg(SPEED_2)
DELTA_SPEED_2 = d(SPEED_2)
DELTA_TIME = d(TIME)
DELTA_DISTANCE = d(DISTANCE)
DELTA_ELEVATION = d(ELEVATION)
DELTA_SPEED = d(SPEED)
DELTA_ENERGY = d(ENERGY)

Power = namedtuple('Power', 'cda, crr, m, p, g')


class PowerStatistics(DataFrameCalculator):

    def _load_data(self, s, ajournal):
        try:
            df = activity_statistics(s, DISTANCE, ELEVATION, SPEED,
                                     activity_journal_id=ajournal.id, with_timespan=True,
                                     log=self._log, quiet=True)
            return add_differentials(df)
        except Exception as e:
            self._log.warning(f'Failed to generate statistics for power: {e}')

    def _extend_data(self, s, df):
        power_ref = self._assert_karg('power')
        power = Power(**loads(Constant.get(s, power_ref).at(s).value))
        self._log.debug('%s: %s' % (power_ref, power))
        df = add_energy_budget(df, power.m, power.g)
        df = add_loss_estimate(df, power.cda, power.crr, power.p)
        df = add_power_estimate(df)
        return df

    def _copy_results(self, s, ajournal, df, loader):
        for time, row in df.iterrows():
            if not pd.isnull(row[POWER]):
                loader.add(POWER, W, AVG, ajournal.activity_group, ajournal, row[POWER], time,
                           StatisticJournalFloat)


def add_differentials(df):
    df[SPEED_2] = df[SPEED] ** 2

    def diff():
        for _, span in df.groupby(TIMESPAN_ID):
            span = span.copy()
            span[TIME] = span.index
            for col in TIME, DISTANCE, ELEVATION, SPEED, SPEED_2:
                span[d(col)] = span[col].diff()
            span[DELTA_TIME] = span[DELTA_TIME] / np.timedelta64(1, 's')
            avg_speed_2 = [(a**2 + a*b + b**2)/3 for a, b in zip(span[SPEED], span[SPEED][1:])]
            span[AVG_SPEED_2] = [np.nan] + avg_speed_2
            yield span

    return pd.concat(diff())


def add_energy_budget(df, m, g=9.8):
    # if DELTA_ELEVATION is +ve we've gone uphill.  so this is the total amount of energy
    # gained in this segment.
    df[DELTA_ENERGY] = m * (df[DELTA_SPEED_2] / 2 + df[DELTA_ELEVATION] * g)
    return df


def add_cda_estimate(df, p=1.225):
    # https://www.cyclingpowerlab.com/CyclingAerodynamics.aspx
    # assume that all energy lost (-ve gain) is due to air resistance.
    df[CDA] = -df[DELTA_ENERGY] / (p * df[AVG_SPEED_2] * df[DELTA_DISTANCE] * 0.5)
    return df


def add_crr_estimate(df):
    # assume that all energy lost is due to rolling resistance
    df[CRR] = -df[DELTA_ENERGY] / df[DELTA_DISTANCE]
    return df


def add_loss_estimate(df, cda=0.45, crr=0, p=1.225):
    # this is the energy spent on air and rolling resistance
    df[LOSS] = (cda * p * df[AVG_SPEED_2] * 0.5 + crr) * df[DELTA_DISTANCE]
    return df


def add_power_estimate(df):
    # power input must balance the energy budget.
    df[POWER] = (df[DELTA_ENERGY] + df[LOSS]) / df[DELTA_TIME]
    return df
